/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */

package simbad.holodeck;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import javax.swing.JDesktopPane;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import simbad.holodeck.HoloManager;
import simbad.sim.*;
import simbad.gui.*;
import javax.swing.*;          

/**
 * This is the Simbad application mainframe.
 *  
 */
public class SimbadHolodeck extends JFrame implements ActionListener{
    static final String version="1.3.1";
    JMenuBar menubar;
    JDesktopPane desktop;
    static int SIZEX = 1500;
    static int SIZEY = 1100;
    WorldWindow worldWindow=null;
    ControlWindow controlWindow=null;
    World world;
    Simulator simulator;
    Console console=null;
    ArrayList agentInspectors=null;
    boolean backgroundMode;
    
    /** Construct Simbad application with the given environment description */
    public SimbadHolodeck(EnvironmentDescription ed, boolean backgroundMode) {
        super("Simbad  - version "+ version);
        this.backgroundMode = backgroundMode;
        JPopupMenu.setDefaultLightWeightPopupEnabled(false);
        try
        {
        	UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
        }
        catch(Exception e)
        {
        	e.printStackTrace();
        }
        desktop = new JDesktopPane();
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setSize(SIZEX, SIZEY);
        createGUI();
        start(ed);
        setVisible(true);
          }
 

    /** Create the main GUI. Only called once.*/
    private void createGUI() {
        getContentPane().add(desktop);
        menubar = new JMenuBar();
        menubar.add(HoloManager.createMenu(this));
        setJMenuBar(menubar);
    }
    
    /** Starts (or Restarts after releaseRessources) the world and simulator.*/
    private void start(EnvironmentDescription ed){
        System.out.println("Starting environmentDescription: "+ed.getClass().getName());
        world = new World(ed);
        simulator = new Simulator(desktop, world, ed);
        createInternalFrames();
        if (backgroundMode) {
            runBackgroundMode();
        }
    }
    
    /** Release all resources. */
    private void releaseRessources(){
        simulator.dispose();
        world.dispose();
        disposeInternalFrames();
    }
    


    /**
     * Creates the windows as Swing InternalFrames
     */
    private void createInternalFrames() {
        worldWindow = new WorldWindow(world);
        desktop.add(worldWindow);
        worldWindow.show();
        worldWindow.setLocation(20, 20); //Place simulator window in top left corner
        agentInspectors = createAgentInspectors(simulator, 820, 0);
        if (!backgroundMode){
        controlWindow = new ControlWindow(world, simulator);
        desktop.add(controlWindow);
        controlWindow.show();
        controlWindow.setLocation(20, 840);
        }
    }
    /**
     * Dispose the windows- used before restart.
     */
    private void disposeInternalFrames() {
        simulator.dispose();
        worldWindow.dispose();
        for(int i = 0; i < agentInspectors.size(); i++)
        	((AgentInspector)agentInspectors.get(i)).dispose();
        if (controlWindow != null){
            controlWindow.dispose();
        }
    }

    /**
     * creates multiple agent inspector windows
     */
    private ArrayList createAgentInspectors(Simulator simulator, int x,
            int y) {
        ArrayList agents = simulator.getAgentList(); //Get the list of agents from the simulator, create local copy
        ArrayList agentInspectors = new ArrayList(agents.size());
        SimpleAgent a;
        for(int i = 0; i < agents.size(); i++){
        	a = ((SimpleAgent) agents.get(i)); //Get each agent from the list of agents
        	if (a instanceof Agent) {
                AgentInspector ai = new AgentInspector((Agent) a, !backgroundMode,simulator);
                desktop.add(ai);
                ai.show();
                ai.setLocation(x+((i+1)*20), y+((i+1)*20));
                agentInspectors.add(i, ai);
            }
        }
        return agentInspectors;
    }
    
    
    public void actionPerformed(ActionEvent event) {
       if (event.getActionCommand() == "open")
       {
    	   releaseRessources();
    	   start(HoloManager.getHoloFromOpenDialogue(event,this));
       }
    }
    
    /** 
     * Runs Simbad in background mode for computation intensive application. 
     * Minimize graphic display and renderer computation.
     */
    private void runBackgroundMode(){
        setTitle(this.getTitle()+" - Background Mode");
        System.out.println("---------------------------------------------");
        System.out.println("Simbad is running in 'Background Mode");
        System.out.println("World is rendered very rarely. UI is disabled");
        System.out.println("--------------------------------------------");
        // slow down the agent inspectors
        for(int i = 0; i < agentInspectors.size(); i++)
        	((AgentInspector)agentInspectors.get(i)).setFramesPerSecond(0.5f);
        // Show a small indication window
        JInternalFrame frame = new JInternalFrame();
        JPanel p = new JPanel();
        p.add(new JLabel("BACKGROUND MODE"));
        frame.setContentPane(p); frame.setClosable(false);
        frame.pack();frame.setLocation(SIZEX/2,SIZEY*3/4);
        desktop.add(frame);
        frame.show();
        world.changeViewPoint(World.VIEW_FROM_TOP,null);
        // start
        simulator.startBackgroundMode();
      
    }
    
    /** The simbad main. Process command line arguments and launch simbad.*/
    public static void main(String[] args) {
        // process options
        boolean backgroundMode = false;
        for (int i = 0; i < args.length;i++)
        {
            if ("-bg".compareTo(args[i])==0) backgroundMode = true;
        }        
        // Check java3d presence
        try {
            Class.forName("javax.media.j3d.VirtualUniverse");
        } catch (ClassNotFoundException e1) {
            JOptionPane.showMessageDialog(null,  "Simbad requires Java 3D", "Simbad 3D",JOptionPane.ERROR_MESSAGE);
            System.exit(-1);
        }

        //request antialising 
        System.setProperty("j3d.implicitAntialiasing", "true");
        SimbadHolodeck frame = new SimbadHolodeck(new simbad.holodeck.Evaluator_Empty() ,backgroundMode);
    }


}