/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.app.Simbad_NEATAvoiderRobot;

import simbad.picoevo.core.representation.*;
import simbad.picoevo.ext.evolution.*;
import simbad.picoevo.ext.representation.*;
import simbad.picoevo.toolbox.*;

import java.io.*;
import java.util.Vector;

public class Main_NEATAvoiderRobot 
{
	static boolean display = true; //false = don't display, true = display
	
	public static void launchAvoiderRobotEvolution()
	{
		Display.info("#### NEAT Test Program (ANN optimisation for the Dynamic Avoidance Problem) ####");

		// 1 - create an Evolution environment along with a single embedded Population
        ParameterSet_Evolution_NEAT parameterSet = new ParameterSet_Evolution_NEAT();
        
        // Set base parameters required by the default Parameter Set
        parameterSet.setInitPopSize(50);
        parameterSet.setGenerations(100);
        parameterSet.setOptimisationFlag(Dictionary.Maximisation);
        
        // Set the parameters required by NEAT
        
    	//Parameters to set for the Compatibility Score between species
    	//Located in Individial_NEATIndividual.getCompatibilityScore
    	parameterSet.setDisjointModifier(1);
    	parameterSet.setExcessModifier(1);
    	parameterSet.setMatchedModifier(0.4);
    	
    	//Parameters to set for adjusting the compatibility threshold
    	//Located in  Population_NEATPopulation_Species.speciate
    	parameterSet.setInitialCompatibilityThreshold(0.5); //this number must NEVER be 0
    	parameterSet.setThresholdIncrement(0.05); //the smaller the number, the more species created
    	parameterSet.setMaximumNumberOfSpecies(30);
    	
    	//Parameters to set for adjusting the fitness level of each individual
    	//Located in Population_NEATPopulation_Species.AdjustFitnesses	
    	parameterSet.setYoungBonusAgeThreshold(10); //if the species are below this age, their fitnesses are boosted
    	parameterSet.setYoungFitnessBonus(1.3); //if the species are young this is how much the fitness is boosted (0.2 = 20%)
    	parameterSet.setOldAgeThreshold(50); //if the species is above this age, its fitness is penalized
    	parameterSet.setOldAgePenalty(0.7); //by this much 
    	
    	//Parameters to set for adjusting the survival rate for spawning
    	//Called in Population_NEATPopulation_SpeciesList::Species::Spawn
    	parameterSet.setSurvivalRate(0.2); //number of population to survive each epoch (0.2 = 20%)
    	
    	//Parameters to control the level of variation 
    	//Located in VariationOperator_Element_NEAT_Gene 
    	//Called in InitialisationOperator_Individual_NEATIndividual during individual initialisation
    	//Called in SelectionOperator_NEAT during mutation 
    	parameterSet.setProbabilityRateReplaced(0.1);
    	parameterSet.setMaxWeightPerturbation(0.5);
    	parameterSet.setActivationMutationRate(0.1);
    	parameterSet.setMaxActivationPerturbation(0.1);
    	
    	//Parameters to control the default genome that is created
    	//Called in InitialisationOperator_Population_NEATPopulation during population initialisation
    	parameterSet.setInitialGenomeInputs(13);
    	parameterSet.setInitialGenomeOutputs(2);

    	//Parameters to control the population
    	//Called in SelectionOperator_NEAT
    	//parameterSet.setPopulationSize();
    	parameterSet.setNumGensAllowedNoImprovement(15);
    	parameterSet.setCrossoverRate(0.07);
    	parameterSet.setMaximumNumberOfNeurons(25); //limits adding more neurons to a spawned child
    	
    	//Parameters to control the amount of variation when adding links to a genome
    	//Used by VariationOperator_Individual_NEAT_MutationAddLink
    	//Used by VariationOperator_Individual_NEAT_MutationAddNode
    	//Called in SelectionOperator_NEAT 
    	parameterSet.setMutationRate(0.3);
    	parameterSet.setChanceAddNode(0.04); //chance of adding a node to a child
    	parameterSet.setChanceAddLink(0.07); //chance of adding a link to a child
    	parameterSet.setChanceOfLoopedLink(0.05); //chance of adding a recurrent link
    	parameterSet.setNumTrysToFindLoop(5); //number of attempts to find a non looped link
    	parameterSet.setNumTrysToAddLink(5); //number of attempts to add a link to a child
    	parameterSet.setNumTrysToFindOldLink(5);
        
        // Done with NEAT Parameters
        String myProblem = new String ("NEATforAvoidanceRobots"); // NO SPACE ALLOWED!
        
        parameterSet.setSelectionOperator(new SelectionOperator_NEAT("NEAT"));
        parameterSet.setEvaluationOperator_Individual(new EvaluationOperator_NEATAvoiderRobot(myProblem,display)); //true = render, false = offscreen render
        parameterSet.setPopulationInitialisationOperator(new InitialisationOperator_Population_NEATPopulation(myProblem));
        parameterSet.setIndividualInitialisationOperator(new InitialisationOperator_Individual_NEAT(myProblem));
        parameterSet.setElementInitialisationOperator(new InitialisationOperator_Element_NEAT_Gene());
        parameterSet.setPopulationStatisticsOperator(new StatisticsOperator_NEAT());
        
        parameterSet.setExperimentPath("Experiments/experiment_"+myProblem+"_"+Misc.getCurrentTimeAsCompactString()+"/");
        parameterSet.check();
        parameterSet.displayInformation();
        
        
        //Let's Evolve!
        Display.info("\n\n## initialization ##");
    	
		World myWorld = new World ("NEATEvolution", parameterSet);
		Population_NEATPopulation neatPop = new Population_NEATPopulation ("NEATpopulation",myWorld);
		myWorld.registerPopulation(neatPop);
		
		neatPop.performInitialisation();

		// 2 - evolution
	
		Display.info("\n\n## evolving ##");
	
		int displayFreq = 10;
		
		for ( int i = 0 ; i != myWorld.getTemplate().getGenerations()+1 ; i++ )
		{
	        if (i % displayFreq == 0)
	        {
	        		myWorld.evolveOneStep(true);
	        }
	        else
	        {
	        		myWorld.evolveOneStep(true);
	        }
		}
		//Log the final information of the experiment
		neatPop.logExperiment();
		Display.info("\n\n\n## terminating ##\n");
	}
	
	public static void launchAvoiderRobotEvolutionWithParams(Vector parameters)
	{
		Display.info("#### NEAT Test Program (ANN optimisation for the Dynamic Avoidance Problem) ####");

		// 1 - create an Evolution environment along with a single embedded Population
        ParameterSet_Evolution_NEAT parameterSet = new ParameterSet_Evolution_NEAT();
        
        // Set base parameters required by the default Parameter Set
        parameterSet.setInitPopSize(((Double)parameters.get(0)).intValue());
        parameterSet.setGenerations(((Double)parameters.get(1)).intValue());
        parameterSet.setOptimisationFlag(Dictionary.Maximisation);
        
        // Set the parameters required by NEAT
        
    	//Parameters to set for the Compatibility Score between species
    	//Located in Individial_NEATIndividual.getCompatibilityScore
    	parameterSet.setDisjointModifier((Double)parameters.get(2));
    	parameterSet.setExcessModifier((Double)parameters.get(3));
    	parameterSet.setMatchedModifier((Double)parameters.get(4));
    	
    	//Parameters to set for adjusting the compatibility threshold
    	//Located in  Population_NEATPopulation_Species.speciate
    	parameterSet.setInitialCompatibilityThreshold((Double)parameters.get(5)); //this number must NEVER be initialised to 0
    	parameterSet.setThresholdIncrement((Double)parameters.get(6)); //the smaller the number, the more species created
    	parameterSet.setMaximumNumberOfSpecies(((Double)parameters.get(7)).intValue());
    	
    	//Parameters to set for adjusting the fitness level of each individual
    	//Located in Population_NEATPopulation_Species.AdjustFitnesses	
    	parameterSet.setYoungBonusAgeThreshold(((Double)parameters.get(8)).intValue()); //if the species are below this age, their fitnesses are boosted
    	parameterSet.setYoungFitnessBonus((Double)parameters.get(9)); //if the species are young this is how much the fitness is boosted (0.2 = 20%)
    	parameterSet.setOldAgeThreshold(((Double)parameters.get(10)).intValue()); //if the species is above this age, its fitness is penalized
    	parameterSet.setOldAgePenalty((Double)parameters.get(11)); //by this much 
    	
    	//Parameters to set for adjusting the survival rate for spawning
    	//Called in Population_NEATPopulation_SpeciesList::Species::Spawn
    	parameterSet.setSurvivalRate((Double)parameters.get(12)); //number of population to survive each epoch (0.2 = 20%)
    	
    	//Parameters to control the level of variation 
    	//Located in VariationOperator_Element_NEAT_Gene 
    	//Called in InitialisationOperator_Individual_NEATIndividual during individual initialisation
    	//Called in SelectionOperator_NEAT during mutation 
    	parameterSet.setProbabilityRateReplaced((Double)parameters.get(13));
    	parameterSet.setMaxWeightPerturbation((Double)parameters.get(14));
    	parameterSet.setActivationMutationRate((Double)parameters.get(15));
    	parameterSet.setMaxActivationPerturbation((Double)parameters.get(16));
    	
    	//Parameters to control the default genome that is created
    	//Called in InitialisationOperator_Population_NEATPopulation during population initialisation
    	parameterSet.setInitialGenomeInputs(((Double)parameters.get(17)).intValue());
    	parameterSet.setInitialGenomeOutputs(((Double)parameters.get(18)).intValue());

    	//Parameters to control the population
    	//Called in SelectionOperator_NEAT
    	//parameterSet.setPopulationSize();
    	parameterSet.setNumGensAllowedNoImprovement(((Double)parameters.get(19)).intValue());
    	parameterSet.setCrossoverRate((Double)parameters.get(20));
    	parameterSet.setMaximumNumberOfNeurons(((Double)parameters.get(21)).intValue()); //limits adding more neurons to a spawned child
    	
    	//Parameters to control the amount of variation when adding links to a genome
    	//Used by VariationOperator_Individual_NEAT_MutationAddLink
    	//Used by VariationOperator_Individual_NEAT_MutationAddNode
    	//Called in SelectionOperator_NEAT 
    	parameterSet.setMutationRate((Double)parameters.get(22));
    	parameterSet.setChanceAddNode((Double)parameters.get(23)); //chance of adding a node to a child
    	parameterSet.setChanceAddLink((Double)parameters.get(24)); //chance of adding a link to a child
    	parameterSet.setChanceOfLoopedLink((Double)parameters.get(25)); //chance of adding a recurrent link
    	parameterSet.setNumTrysToFindLoop(((Double)parameters.get(26)).intValue()); //number of attempts to find a non looped link
    	parameterSet.setNumTrysToAddLink(((Double)parameters.get(27)).intValue()); //number of attempts to add a link to a child
    	parameterSet.setNumTrysToFindOldLink(((Double)parameters.get(28)).intValue());
        
        // Done with NEAT Parameters
        String myProblem = new String ((String)parameters.get(29)); // NO SPACE ALLOWED!
        
        parameterSet.setSelectionOperator(new SelectionOperator_NEAT("NEAT"));
        parameterSet.setEvaluationOperator_Individual(new EvaluationOperator_NEATAvoiderRobot(myProblem,display)); //true = render, false = offscreen render
        parameterSet.setPopulationInitialisationOperator(new InitialisationOperator_Population_NEATPopulation(myProblem));
        parameterSet.setIndividualInitialisationOperator(new InitialisationOperator_Individual_NEAT(myProblem));
        parameterSet.setElementInitialisationOperator(new InitialisationOperator_Element_NEAT_Gene());
        parameterSet.setPopulationStatisticsOperator(new StatisticsOperator_NEAT());
        
        parameterSet.setExperimentPath("Experiments/experiment_"+myProblem+"_"+Misc.getCurrentTimeAsCompactString()+"/");
        parameterSet.check();
        parameterSet.displayInformation();
        
        
        //Let's Evolve!
        Display.info("\n\n## initialization ##");
    	
		World myWorld = new World ("NEATEvolution", parameterSet);
		Population_NEATPopulation neatPop = new Population_NEATPopulation ("NEATpopulation",myWorld);
		myWorld.registerPopulation(neatPop);
		
		neatPop.performInitialisation();

		// 2 - evolution
	
		Display.info("\n\n## evolving ##");
	
		int displayFreq = 10;
		
		for ( int i = 0 ; i != myWorld.getTemplate().getGenerations()+1 ; i++ )
		{
			//Individual_NEATIndividual best = null;
	        if (i % displayFreq == 0)
	        {
	        		myWorld.evolveOneStep(true);
	        }
	        else
	        {
	        		myWorld.evolveOneStep(true);
	        }
		}
		//Log the final information of the experiment
		neatPop.logExperiment();
		Display.info("\n\n\n## terminating ##\n");
	}
	
	public static void main(String[] args) {
		double startTime = System.currentTimeMillis();
		String filename = new String();
		Vector parameters = null;
        Version.displayNeuroevolutionarySolverReleaseInformation();
		
        if(args.length == 1)
        {
        	try
        	{
        		//open the parameter file and prep for tokenizing
        		filename = args[0];
        		BufferedReader in = new BufferedReader(new FileReader(filename));  
        		StreamTokenizer token = new StreamTokenizer(in);
        		parameters = new Vector();
        		
        		//tokenize the parameter file
        		while(token.nextToken() != StreamTokenizer.TT_EOF) 
        			   switch(token.ttype) 
        			   {
        			    case StreamTokenizer.TT_WORD:
        			        parameters.add(token.sval);
        			        break;
        			    case StreamTokenizer.TT_NUMBER:
        			        parameters.add(new Double(token.nval));
        			        break;
        			   }
        	}
        	catch(IOException e)
        	{
	    		e.printStackTrace();
	    		Display.error("Main_NEATAvoiderRobot::main(.) - cannot open parameter file [" + filename + "]");
        	}
        }
        else
        {
        	parameters = new Vector();
        }
        //Check to be certain we have the right amount of parameters. They MUST be in the correct order.
    	if(parameters.size() == 30)
    	{
    		Display.info("Parameter filename is: " + filename);
            Display.info("### Neuroevolutionary Solver - NEAT Avoider Robot ###");
            Display.info("Running...");
    		launchAvoiderRobotEvolutionWithParams(parameters);
    		Display.info("\nTerminated ("+ ((System.currentTimeMillis()-startTime)/1000) +"s elapsed).");
    		System.exit(0);
    	}
    	else
    	{
            Display.info("### Neuroevolutionary Solver - NEAT Avoider Robot ###");
            Display.info("Running...");        
            launchAvoiderRobotEvolution(); 	
            Display.info("\nTerminated ("+ ((System.currentTimeMillis()-startTime)/1000) +"s elapsed).");
            System.exit(0);
    	}
	}
}
