/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.app.Simbad_SDNEATAvoiderRobot;

import java.util.ArrayList;

import javax.vecmath.Vector3d;

import simbad.sim.Agent;
import simbad.sim.LampActuator;
import simbad.sim.RangeSensorBelt;
import simbad.sim.RobotFactory;

public class NeuralAgent extends Agent {

	//RangeSensorBelt bumpers;
	RangeSensorBelt sonars;
	//Lamp for visual inspection of agent
	LampActuator lamp;
	ArrayList<Double> inputControllerValuesList;
	Vector3d startPosition;
	Vector3d goal;
	Vector3d currentPosition;
	int numOfCollisions;
	double d; //distance to goal calculated by pythagoras.
	private Evaluator_Environment parent;
	
	public NeuralAgent(Vector3d position, Vector3d givenGoal, String name, Evaluator_Environment theParent) {
    	super(position, name);

    	//initialise
    	parent = theParent;
        inputControllerValuesList = new ArrayList<Double>();
        startPosition = new Vector3d(position);
        goal = new Vector3d(givenGoal);
        currentPosition = new Vector3d();
        d = 0.0;
        numOfCollisions = 0;
        // Add sensors
        sonars = RobotFactory.addSonarBeltSensor(this,12);
        lamp = RobotFactory.addLamp(this);
    }
	
	/** Initialize Agent's Behavior -- launched before the first performOneStep */
    public void initBehavior() {
    }
    
    /** move robot at start position and load the new genome  */
    public void resetEvaluation() {
    	
		this.moveToStartPosition();
	
		this.parent.fitness = 0;
		this.parent.isRunnable = true;
    }
    
    public int getCollisionCount()
    {
    	return numOfCollisions;
    }
    
    public void resetCollisionCount()
    {
    	numOfCollisions = 0;
    }
    
    public double getDistanceToGoal()
    {
    	return d;
    }
    
    public void performBehavior()
    {
    	// Computing the fitness
    	// step 1 of 2
    	// updating the fitness for the previous step and checking for collision
    	//Control code for the overhead lamp on the neural agent. Shows collisions and incoming collisions
    	lamp.setBlink(false);
    	lamp.setOn(false);
    	if (sonars.oneHasHit()) {
    		lamp.setBlink(true);
    		// reads the three front quadrants
    		double left = sonars.getFrontLeftQuadrantMeasurement();
    		double right = sonars.getFrontRightQuadrantMeasurement();
    		double front = sonars.getFrontQuadrantMeasurement();
    		// if obstacle near
    		if ((front  < 1)||(left  < 1)||(right  < 1)) {
    			lamp.setBlink(false);
    			lamp.setOn(true);
    		}
    	}
    	else if(collisionDetected() == true)
    	{
    		lamp.setOn(true);
    	}

    	//record the collision in a running total
    	if(collisionDetected() == true)
    	{
    		numOfCollisions++;
    	}

    	// * reads the sonar values

    	double[] inputValue = new double[sonars.getNumSensors()];

    	for( int i = 0; i < sonars.getNumSensors(); i++)
    	{
    		inputValue[i] = sonars.getMeasurement(i);
    		//Make the sensor values manageable numbers
    		if ( inputValue[i] == Double.POSITIVE_INFINITY)
    		{
    			inputValue [i] = 1.5;
    		}
    		//A very low percentage of sonar inputs will receive noise
    		//This will either increase or decrease their sonar readings by 0.01 - 0.09 metres
    		//roughly 9 centimetres or less inaccuracy
    		if ( Math.random() > 0.50 )
    		{
    			if ( Math.random() < 0.01 )
    			{
    				inputValue [i] += Math.random()/100;
    			}
    		}
    		else
    		{
    			if ( Math.random() < 0.01 )
    			{
    				inputValue [i] -= Math.random()/100;
    			}
    		}	
    	}

    	getCoords(currentPosition); 
    	//Calculate distance to goal using Pythagoras. 
    	d = Math.sqrt(Math.pow((currentPosition.x - goal.x), 2.0)+Math.pow((currentPosition.z - goal.z), 2.0));

    	// * load the new input values into the controller
    	inputControllerValuesList.clear();
    	for ( int i = 0 ; i < sonars.getNumSensors() ; i++ )
    	{
    		inputControllerValuesList.add( new Double(inputValue[i]) );
    	}

    	//add the distance to goal as the last input value
    	inputControllerValuesList.add(new Double(d));

    	//run the values through the neural network
    	parent.phenome.step(inputControllerValuesList);

    	// * perform locomotion step
    	double speed = parent.phenome.getOutputNeuronAt(0).getOutputValue();
    	double angle = parent.phenome.getOutputNeuronAt(1).getOutputValue();
    	setTranslationalVelocity( speed );
    	setRotationalVelocity( angle );

    	// Computing the fitness
    	// step 2 of 2
    	// penalize if not moving
    	double maxSensorValue = 0;
    	for ( int i = 0 ; i != sonars.getNumSensors() ; i++ )
    		if ( maxSensorValue < inputValue[i] )
    			maxSensorValue = inputValue[i];

    	if (d < 0.5)
    	{
    		setTranslationalVelocity( 0.0 );
    		setRotationalVelocity( 0.0 );
    		parent.fitness += 5.0;
    	}
    	else
    	{
    		parent.fitness += speed*(1.0 - Math.abs(angle)) * ( 1.6 - maxSensorValue )*(1.0/d);
    	}
    }
}
