/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.ext.evolution;

import java.util.*;

import simbad.picoevo.core.evolution.*;
import simbad.picoevo.core.representation.*;
import simbad.picoevo.ext.representation.*;
import simbad.picoevo.toolbox.*;
import simbad.picoevo.toolbox.Dictionary;

public class SelectionOperator_NEAT extends SelectionOperator {

	Random rand;
	
	public SelectionOperator_NEAT() {
		super._name = "NEAT speciate and spawn selection operator";
		rand = new Random();
	}

	public SelectionOperator_NEAT(String __name) {
		super(__name);
		rand = new Random();
	}

	public void performSelection ( PopulationObject populationList )
	{
		Population_NEATPopulation population = (Population_NEATPopulation)populationList;

		//In the original version of NEAT they would check the incoming vector of fitness scores
		//for the correct amount right here, we don't need to do this since we are dealing with populations
		//that retain their own fitness scores.
		
		//First we reset appropriate values from the population and any poorly performing species
		population = resetAndKill(population);
		
		//Now we would update the genome with the fitnesses scored in the last run
		//Again we don't have to do this. These raw fitness values are already available in the population
		
		//sort genomes and keep a record of the best performer
		population = sortAndRecord(population);
		
		//separate the population into species of similar topology, adjust fitnesses and calculate spawn levels
		population.speciesList.speciate(populationList.getIndividualList());
		
		//once speciation occurs, we can log the genomes
		population.logGeneration("Generation" + population.getGenerationNumber());
		
		//this will hold the new population of genomes
		PopulationContainer newpop = new PopulationContainer();
		
		//request the offspring from each species. The number of children to spawn is a double
		//which we need to convert to an int
		int numSpawnedSoFar = 0;
		
		//the latest new offspring to deal with on this iteration
		Individual_NEATIndividual offspring = null;
		
		for(int species = 0; species < population.speciesList.numOfSpecies(); species++)
		{
			if(numSpawnedSoFar < population.getTemplate().getInitPopSize())
			{
				//This is the amount of offspring this species is required to spawn
				long numToSpawn = Math.round(population.speciesList.getSpeciesAt(species).spawnsRequired);
				
				boolean chosenBestYet = false;
				
				while(numToSpawn-- > 0)
				{
					//first grab the best performing genome from this species and transfer
					//to the new population without mutation. This provides per species elitism
					if(!chosenBestYet)
					{
						offspring = population.speciesList.getSpeciesAt(species).leader();
						chosenBestYet = true;
					}
					//if greater than one we can use the crossover operator
					else
					{
						//spawn 1
						Individual_NEATIndividual parent1 = population.speciesList.getSpeciesAt(species).spawn();
						
						if(Math.random() < ((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getCrossoverRate())
						{
							//spawn 2 and make sure it's not the same as spawn 1
							Individual_NEATIndividual parent2 = population.speciesList.getSpeciesAt(species).spawn();
							//number of attempts at finding a different genome
							int numAttempts = 5;
							while (parent1.genomeID == parent2.genomeID && numAttempts-- > 0)
							{
								parent2 = population.speciesList.getSpeciesAt(species).spawn();
							}
							if (parent1.genomeID != parent2.genomeID)
							{
								VariationOperator_Population_NEAT_CrossOver op = new VariationOperator_Population_NEAT_CrossOver("crossover");
								op.applyOperator(population,parent1,parent2);
							}
						}
						else
						{
							offspring = parent1;
						}
					
					
						//give the offspring its own ID
						population.nextGenomeID++;
						offspring.setGenomeID(population.nextGenomeID);
					
						//now we have a spawned child lets mutate it!
						//First there is the chance a neuron may be added
						if(offspring.getNumNeurons() < ((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getMaximumNumberOfNeurons())
						{
							//setup a node addition mutation
							VariationOperator_Individual_NEAT_MutationAddNode op = 
								new VariationOperator_Individual_NEAT_MutationAddNode("Add node mutation",
																				  	((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getChanceAddNode(),
																				  	population.innovationList,
																				  	((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getNumTrysToFindOldLink());
							offspring.addOperator(op);//add a mutation to this offspring
						}
					
						//now there is a chance a link may be added
						VariationOperator_Individual_NEAT_MutationAddLink op2 =
							new VariationOperator_Individual_NEAT_MutationAddLink("Add Link mutation",
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getChanceAddLink(),
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getChanceOfLoopedLink(),
									population.innovationList,
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getNumTrysToFindLoop(),
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getNumTrysToAddLink());
					
						offspring.addOperator(op2);
					
						//mutate the weights
						VariationOperator_Element_NEAT_Gene op3 =
							new VariationOperator_Element_NEAT_Gene("Mutate Weights",
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getMutationRate(),
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getProbabilityRateReplaced(),
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getMaxWeightPerturbation(),
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getActivationMutationRate(),
									((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getMaxActivationPerturbation());
					
						for(int i = 0; i < offspring.getElementListSize(); i++)
						{
							offspring.getElementAt(i).addOperator(op3);
						}
					
						//now fire off all the queued up mutations to this offspring all at once.
						offspring.performVariations();
					}
					//sort the offsprings genes by their innovation numbers.
					//greatest to least
					offspring.sort();
					
					//lets clone this offspring and add it to the new population
					newpop.cloneAndRegisterIndividual(offspring);
					
					//increase the number spawned for this species
					numSpawnedSoFar++;
					
					if (numSpawnedSoFar == ((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getInitPopSize())
					{
						numToSpawn = 0;
					}
				}//end while
			}//end if
		}//next species
		
		//If there is an underflow due to the rounding error and the amount of offspring
		//falls short of the population size. Additional children need to be created and 
		//added to the new population. This is achieved simply by using tournament selection
		//over the entire population.
		if (numSpawnedSoFar < ((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getInitPopSize())
		{
			if(numSpawnedSoFar == 0)
			{
				Display.error(""+this.getClass().getName()+"::performSelection - underflow is the entire population, modify fitness function so it doesn't always return 0");
			}
			int requiredChildren = ((ParameterSet_Evolution_NEAT)population.getOwner().getTemplate()).getInitPopSize() - numSpawnedSoFar;
			
			//grab them
			while(requiredChildren-- > 0)
			{
				newpop.cloneAndRegisterIndividual(tournamentSelection(population, population.populationSize/5));
			}
		}

		//replace the current population with the new one
		//also increases generation number
		((Population)populationList).renewPopulation(newpop);
		population = null;
	}
	
	//returns a random integer between x and y
	private int randInt(int x, int y)
	{
		return rand.nextInt(Integer.MAX_VALUE)%(y-x+1)+x;
	}
	
	//Tournament selection
	private Individual_NEATIndividual tournamentSelection(Population_NEATPopulation population, int numComparisons)
	{
		double bestFitnessSoFar = 0.0;
		int chosenOne = 0;
		
		for(int i = 0; i < numComparisons; i++)
		{
			int thisTry = randInt(0, population.getIndividualList().size()-1);
			
			if(((Individual_NEATIndividual)population.getIndividual(i)).getFitness() > bestFitnessSoFar)
			{
				chosenOne = thisTry;
				bestFitnessSoFar = ((Individual_NEATIndividual)population.getIndividual(i)).getFitness();
			}
		}
		
		//return the champion
		return (Individual_NEATIndividual)population.getIndividual(chosenOne);
	}
	
	//perform a little cleanup, sort the population into descending fitness, and keep a record
	//of the best genome
	private Population_NEATPopulation sortAndRecord(Population_NEATPopulation pop)
	{
		//sort the genomes according to their unadjusted (no fitness sharing) fitnesses
		pop.sort();
		
		//is the best genome this generation the best ever?
		if( pop.getTemplate().getOptimisationFlag() == Dictionary.Maximisation )
		{
			if(((Individual_NEATIndividual)(pop.getIndividual(0))).getRawFitness() > pop.bestFitnessEver);
			{
				pop.bestFitnessEver = ((Individual_NEATIndividual)(pop.getIndividual(0))).getRawFitness();
			}
		}
		else
		{
			if(((Individual_NEATIndividual)(pop.getIndividual(pop.getPopulationSize()-1))).getRawFitness() > pop.bestFitnessEver);
			{
				pop.bestFitnessEver = ((Individual_NEATIndividual)(pop.getIndividual(pop.getPopulationSize()-1))).getRawFitness();
			}
		}
		
		return pop;
	}
	
	//This function resets some values for the next population and kills off any poorly performing species
	private Population_NEATPopulation resetAndKill( Population_NEATPopulation pop)
	{
		pop.totalFitnessAdjusted = 0;
		pop.averageFitnessAdjusted = 0;
		
		//purge the species
		int currentSpeciesNumber = 0;
		
		while(currentSpeciesNumber != pop.speciesList.numOfSpecies())
		{
			pop.speciesList.getSpeciesAt(currentSpeciesNumber).purge();
			
			//kill off species if not improving and if not the species
			//which contains the best genome found so far
			if(((pop.speciesList.getSpeciesAt(currentSpeciesNumber).gensNoImprovement) > 
			   ((ParameterSet_Evolution_NEAT)pop.getTemplate()).getNumGensAllowedNoImprovement()) &&
			   (pop.speciesList.getSpeciesAt(currentSpeciesNumber).bestFitness < pop.bestFitnessEver))
				{
					pop.speciesList.eraseSpeciesAt(currentSpeciesNumber);
					currentSpeciesNumber--;
				}
			currentSpeciesNumber++;
		}
		return pop;
	}
}
