/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.ext.evolution;

import java.util.*;

import simbad.picoevo.core.evolution.*;
import simbad.picoevo.core.representation.*;
import simbad.picoevo.ext.representation.*;
import simbad.picoevo.toolbox.*;

public class VariationOperator_Individual_NEAT_MutationAddNode extends VariationOperator {

	//rate from VariationOperator will correspond to mutationRate
	protected double mutationRate;
	protected Population_NEATPopulation_InnovationList innovationList;
	protected int numTrysToFindOldLink;
	//create a random number generator
	Random rand;
	
	public VariationOperator_Individual_NEAT_MutationAddNode(String __name) {
		super(__name, 0.0);
		mutationRate = 0.0;
		innovationList = null;
		numTrysToFindOldLink = 1;
		rand = new Random();
	}

	public VariationOperator_Individual_NEAT_MutationAddNode(String __name, double rate) {
		super(__name, rate);
		mutationRate = rate;
		innovationList = null;
		numTrysToFindOldLink = 1;
		rand = new Random();
	}
	
	public VariationOperator_Individual_NEAT_MutationAddNode(String __name, Population_NEATPopulation_InnovationList _innovationList,
															int _numTrysToFindOldLink) {
		super(__name, 0.0);
		mutationRate = 0.0;
		innovationList = _innovationList;
		numTrysToFindOldLink = _numTrysToFindOldLink;
		rand = new Random();
	}

	public VariationOperator_Individual_NEAT_MutationAddNode(String __name, double rate, Population_NEATPopulation_InnovationList _innovationList,
															int _numTrysToFindOldLink) {
		super(__name, rate);
		mutationRate = rate;
		innovationList = _innovationList;
		numTrysToFindOldLink = _numTrysToFindOldLink;
		rand = new Random();
	}

	@Override
	public void applyOperator(EvolutionObject object) {

		Individual_NEATIndividual genome = (Individual_NEATIndividual)object;

		//just return if we are not mutating
		if(Math.random() > mutationRate)
		{
			return;
		}
		
		//if a valid link is found into which to insert the new neuron, this is set to true
		boolean insert = false;
		
		//this will hold the index into our vector of links to the chosen link
		int chosenLink = 0;
		
		//First a link is chosen to split. If the genome is small the code makes sure one of the older
		//links is split to ensure a chaining effect does not occur. Here if the genome contains less than
		//5 hidden neurons it is considered to be too small to select a link at random
		int sizeThreshold = genome.numInputs() + genome.numOutputs() + 10;
		
		if(genome.getNumLinks() < sizeThreshold)
		{
			while (numTrysToFindOldLink-- > 0)
			{
				//choose a link with a bias towards the older links in the genome
				int upperValue = genome.getNumLinks()-1-(int)Math.sqrt((double)genome.getNumLinks());
				chosenLink = randInt(0,upperValue);
				
				//make sure the link is enabled and that it is not a recurrent link
				//or has a bias input
				int fromNeuron = genome.getLinkAt(chosenLink).getFromNeuron();
				if(genome.getLinkAt(chosenLink).isEnabled() &&
				   !genome.getLinkAt(chosenLink).getRecurrent() &&
				   genome.getNeuronAt(getElementPosition(fromNeuron,genome)).getNeuronType() != 4)
				{
					insert = true;
					numTrysToFindOldLink = 0;
				}
			}
			if(!insert)
			{
				//failed to find a decent link
				return;
			}
		}
		else
		{
			while(!insert)
			{
				chosenLink = randInt(0,genome.getNumLinks()-1);
				
				//make sure the link is enabled and that it is not a recurrent link or has a bias input
				int fromNeuron = genome.getLinkAt(chosenLink).getFromNeuron();
				if(genome.getLinkAt(chosenLink).isEnabled() &&
				   !genome.getLinkAt(chosenLink).getRecurrent() &&
				   genome.getNeuronAt(getElementPosition(fromNeuron,genome)).getNeuronType() != 4)
				{
					insert = true;
				} 
			}
		}
		
		//disable this gene
		genome.getLinkAt(chosenLink).disable();
		
		//grab the weight from the gene (we want to use this for the weight of one of the new links
		//so the split does not disturb anything the NN may have already learned.
		double originalWeight = genome.getLinkAt(chosenLink).getLinkWeight();
		
		//identify the neurons this link connects
		int from = genome.getLinkAt(chosenLink).getFromNeuron();
		int to = genome.getLinkAt(chosenLink).getToNeuron();
		
		//calculate the depth and width of the new neuron. We can use the depth to see if the link
		//feeds backward or forwards
		
		double newDepth = (genome.getNeuronAt(getElementPosition(from,genome)).getSplitY() +
						  genome.getNeuronAt(getElementPosition(to,genome)).getSplitY()) /2;
		double newWidth = (genome.getNeuronAt(getElementPosition(from,genome)).getSplitX() +
						  genome.getNeuronAt(getElementPosition(to,genome)).getSplitX()) /2;
		
		//now to see if this innovation has been created previously by another member of the population
		//innovation type is 0 since we are adding a new node
		int id = innovationList.checkInnovation(from, to, 0);
		
		/*it is possible for NEAT to repeatedly do the following:
		  
	      1. Find a link. Lets say we choose link 1 to 5
	      2. Disable the link,
	      3. Add a new neuron and two new links
	      4. The link disabled in Step 2 may be re-enabled when this genome
	         is recombined with a genome that has that link enabled.
	      5  etc etc

	  	Therefore, this function must check to see if a neuron ID is already 
	  	being used. If it is then the function creates a new innovation
	  	for the neuron. */
		if(id >= 0)
		{
			int neuronId = innovationList.getNeuronID(id);
			if(alreadyHaveThisNeuronID(neuronId,genome))
			{
				id = -1;
			}
		}
		
		if(id < 0)
		{
			//1. Create the new innovation for the new neuron
			//Innovation type, 0 = node, 1 = link
			//Node type, 0=none, 1=input, 2=output, 3=hidden, 4=bias
			int idNeuron1 = innovationList.getNextInnovationID();
			int newNeuronID = innovationList.createNewInnovation(from, to, 0, 3, newWidth, newDepth);
			
			//2. Create a new gene
			//start by creating a variation operator for randomisation and one to do nothing
			VariationOperator_Element_NEAT_Gene op = new VariationOperator_Element_NEAT_Gene("initialization operator",
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getMutationRate(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getProbabilityRateReplaced(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getMaxWeightPerturbation(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getActivationMutationRate(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getMaxActivationPerturbation());
			
			VariationOperator_Element_NEAT_Gene op2 = new VariationOperator_Element_NEAT_Gene("initialization operator - does nothing");
			
			Element_NEAT_NGene node = new Element_NEAT_NGene("mutation node",
															 genome, //owner is this genome
															 op, //initialise with this operation
															 idNeuron1, //new innovation id
															 false, //is not recurrent ?
															 3, //neuron type is hidden
															 0.0, //activation response for new node
															 newWidth, //splitX
															 newDepth, //splitY
															 newNeuronID //new unique neuronID
															 );
			genome.addNeuron(node);
			
			//3. Create the first new link
			int idLink1 = innovationList.getNextInnovationID();
			//create the new innovation for the first link
			innovationList.createNewInnovation(from, newNeuronID, 1);
			//create the new link gene
			Element_NEAT_LGene nLink1 = new Element_NEAT_LGene("mutation link for node",
					  genome,
					  op2, //this initialisation does nothing
					  idLink1, //this is getting the next neuron id it should be innovation id
					  false, //not recurrent?
					  from, //from the old neuron
					  newNeuronID, //to the new neuron
					  1.0); //set this vector weight to 1.0 to not affect NN learning
			genome.addLink(nLink1);
			
			//4. Create the second new link
			int idLink2 = innovationList.getNextInnovationID();
			//create the new innovation for the first link
			innovationList.createNewInnovation(newNeuronID, to, 1);
			//create the new link gene
			Element_NEAT_LGene nLink2 = new Element_NEAT_LGene("mutation link for node",
					  genome,
					  op2, //this initialisation does nothing
					  idLink2, //this is getting the next neuron id it should be innovation id
					  false, //not recurrent?
					  newNeuronID, //from the old neuron
					  to, //to the new neuron
					  originalWeight); //set this vector weight to the original value to not affect learning
			genome.addLink(nLink2);
		}
		else
		{
			//this innovation has already been created so grab the relevant neuron
			//and link info from the innovation database
			int newNeuronID = innovationList.getNeuronID(id);
			
			//get the innovation IDs for the two new link genes.
			int idLink1 = innovationList.checkInnovation(from, newNeuronID, 1);
			int idLink2 = innovationList.checkInnovation(newNeuronID, to, 1);
			
			//this should never happen because the innovations "should" have already occurred
			if( idLink1 < 0 || idLink2 < 0)
			{
				Display.error(""+this.getClass().getName()+"::mutate(.) - Attempting to a link in a new non-innovated node which is missing link innovations");
			}
			
			//Create the new neuron
			//start by creating a variation operator for randomisation and one to do nothing
			VariationOperator_Element_NEAT_Gene op = new VariationOperator_Element_NEAT_Gene("initialization operator",
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getProbabilityRateReplaced(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getProbabilityRateReplaced(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getMaxWeightPerturbation(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getActivationMutationRate(),
																							((ParameterSet_Evolution_NEAT)genome.getTemplate()).getMaxActivationPerturbation());
			
			VariationOperator_Element_NEAT_Gene op2 = new VariationOperator_Element_NEAT_Gene("initialization operator - does nothing");
			
			//Create the neuron and add it
			Element_NEAT_NGene node = new Element_NEAT_NGene("mutation node",
															 genome, //owner is this genome
															 op, //initialise with this operation
															 id,/*0,*/ //should be an existing innovation id
															 false, //is not recurrent ?
															 3, //neuron type is hidden
															 0.0, //activation response for new node
															 newWidth, //splitX
															 newDepth, //splitY
															 newNeuronID //new unique neuronID
			);
			genome.addNeuron(node);

			//now we need to create two new genes to represent the new links
			
			//3. Create the first new link
			//create the new link gene
			Element_NEAT_LGene nLink1 = new Element_NEAT_LGene("mutation link for node",
					  genome,
					  op2, //this initialisation does nothing
					  idLink1, //this is getting the next neuron id it should be innovation id
					  false, //not recurrent?
					  from, //from the old neuron
					  newNeuronID, //to the new neuron
					  1.0); //set this vector weight to 1.0 to not affect NN learning
			genome.addLink(nLink1);
			
			//4. Create the second new link
			//create the new link gene
			Element_NEAT_LGene nLink2 = new Element_NEAT_LGene("mutation link for node",
					  genome,
					  op2, //this initialisation does nothing
					  idLink2, //this is getting the next neuron id it should be innovation id
					  false, //not recurrent?
					  newNeuronID, //from the old neuron
					  to, //to the new neuron
					  originalWeight); //set this vector weight to the original value to not affect learning
			genome.addLink(nLink2);
		}
		return;
	}
	
	//tests to see if the given neuron id exists in the given genome
	//returns true if it exists
	private boolean alreadyHaveThisNeuronID(int id, Individual_NEATIndividual genome)
	{
		for(int neuron = 0; neuron < genome.getNumNeurons(); neuron++)
		{
			if(id == genome.getNeuronAt(neuron).getNeuronID())
			{
				return true;
			}
		}
		return false;
	}
	
	//given a neuron ID this function finds its position in the vector of neurons
	private int getElementPosition(int neuronId, Individual_NEATIndividual genome)
	{
		for(int neuron = 0; neuron < genome.getNumNeurons(); neuron++)
		{
			if(genome.getNeuronAt(neuron).getNeuronID() == neuronId)
			{
				return neuron;
			}
		}
		Display.error(""+this.getClass().getName()+"::mutate(.) - Attempting to a link to a node which apparently doesn't exist in this genome");
		return -1;
	}
	
	//returns a random integer between x and y
	private int randInt(int x, int y)
	{
		return rand.nextInt(Integer.MAX_VALUE)%(y-x+1)+x;
	}

}
