/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.ext.evolution;

import java.util.*;

import simbad.picoevo.core.evolution.*;
import simbad.picoevo.core.representation.*;
import simbad.picoevo.ext.representation.*;
import simbad.picoevo.toolbox.*;

public class VariationOperator_Individual_SDNEAT_MutationSegmentDuplicate extends VariationOperator {

	//rate from VariationOperator will correspond to mutationRate
	protected double mutationRateSD;
	protected Population_NEATPopulation_InnovationList innovationList;
	protected int maxSDMutationAttempt; //number of attempts to find a full path
	protected boolean abortSD; //true when this SD attempt should be aborted.
	protected boolean completeSD; //true when a proper path has been found
	//create a random number generator
	Random rand;
	
	public VariationOperator_Individual_SDNEAT_MutationSegmentDuplicate(String __name) {
		super(__name, 0.0);
		mutationRateSD = 0.0;
		innovationList = null;
		maxSDMutationAttempt = 1;
		abortSD = false;
		completeSD = false;
		rand = new Random();
	}

	public VariationOperator_Individual_SDNEAT_MutationSegmentDuplicate(String __name, double rate) {
		super(__name, rate);
		mutationRateSD = rate;
		innovationList = null;
		maxSDMutationAttempt = 1;
		abortSD = false;
		completeSD = false;
		rand = new Random();
	}
	
	public VariationOperator_Individual_SDNEAT_MutationSegmentDuplicate(String __name, Population_NEATPopulation_InnovationList _innovationList,
															int _maxSDMutationAttempt) {
		super(__name, 0.0);
		mutationRateSD = 0.0;
		innovationList = _innovationList;
		maxSDMutationAttempt = _maxSDMutationAttempt;
		abortSD = false;
		completeSD = false;
		rand = new Random();
	}

	public VariationOperator_Individual_SDNEAT_MutationSegmentDuplicate(String __name, double rate, Population_NEATPopulation_InnovationList _innovationList,
															int _maxSDMutationAttempt) {
		super(__name, rate);
		mutationRateSD = rate;
		innovationList = _innovationList;
		maxSDMutationAttempt = _maxSDMutationAttempt;
		abortSD = false;
		completeSD = false;
		rand = new Random();
	}

	/** Perform Segmental Duplication
	 * 
	 * Definition: A segment contains n nodes, where n is greater than zero and n does not contain input or output nodes.
	 * The segments first link must come from an input node and its last link must lead to an output node.
	 * A segment does not include recurrent links.
	 * 
	 * Behaviour: This function will attempt to find a path in the network from a random input node to a random
	 * output node. Once a path is found the nodes in that path will be duplicated and inserted as new
	 * innovations along with a set of new links connecting it to the original input and output node. The values from
	 * the original list of nodes should have been copied to the new segment to preserve the original knowledge
	 * but the segment is subject to higher than normal mutation rates as it is hoped the segment might make
	 * a generational quantum leap.
	 * @param object
	 */
	@Override
	public void applyOperator(EvolutionObject object) {

		//The individual we may perform segmental duplication on.
		Individual_NEATIndividual genome = (Individual_NEATIndividual)object;
		//The nodes we are going to duplicate during this segmental duplication.
		//neuronType 0=none, 1=input, 2=output, 3=hidden, 4=bias
		Vector<Element_NEAT_NGene> nodesToDuplicate = new Vector<Element_NEAT_NGene>();
		//The links we are going to duplicate during this segmental duplication.
		Vector<Element_NEAT_LGene> linksToDuplicate = new Vector<Element_NEAT_LGene>();
		
		//Just return if we are not mutating
		if(Math.random() > mutationRateSD)
		{
			return;
		}
		
		//If there are no hidden nodes, we should not attempt a segmental duplication.
		//We cannot have doubly linked nodes.
		int actualNodeAmount = genome.getNeurons().size();
		int inputOutputAmount = genome.numInputs()+genome.numOutputs();
		if(actualNodeAmount > inputOutputAmount)
		{
			//First lets find a sequence to perform segemental duplication on
			while(maxSDMutationAttempt-- > 0)
			{
				//Choose a random input node
				nodesToDuplicate.add(genome.getNeuronAt(randInt(0,genome.inputs-1)));
				//Make sure we got an input node, abort otherwise
				if(nodesToDuplicate.elementAt(0).getNeuronType() != 1)
				{
					abortSD = true;
				}
				else
				{
					while(nodesToDuplicate.elementAt(nodesToDuplicate.size()-1).getNeuronType() != 2 && !abortSD)
					{
						//find a random link from the last node that isn't recurrent or disabled
						//add that link and its connected neuron to the links and nodes to duplicate
						linksToDuplicate.add(getPathNextRandomNode(nodesToDuplicate.elementAt(nodesToDuplicate.size()-1),genome));
						int nextNodePosition = getElementPosition(linksToDuplicate.elementAt(linksToDuplicate.size()-1).getToNeuron(),genome);
						nodesToDuplicate.add(genome.getNeuronAt(nextNodePosition));
					}
					//Now check to make sure we haven't gone from the input node directly to an output
					//if we have, abort this attempt.
					if(nodesToDuplicate.size() == 2 && nodesToDuplicate.elementAt(nodesToDuplicate.size()-1).getNeuronType() == 2)
					{
						//abort, this is a single link from input to output, can't segmentally duplicate it
						abortSD = true;
					}
				}
				
				if(abortSD)
				{
					//Somewhere the path did not complete and we will abort this attempt
					nodesToDuplicate.clear();
					linksToDuplicate.clear();
					abortSD = false;
				}
				else
				{
					//We found a valid path and have recorded the links and nodes
					maxSDMutationAttempt = 0;
					completeSD = true;
				}
			}
			//Now we have a complete sequence for segmental duplication
			//let's duplicate it
			//Two cases:
			// 1. There is only one hidden node, in which case we treat this similarly to inserting a new node
			// 2. There is more than one hidden node and we have to deal with knitting them together.
			if(completeSD)
			{
				//counters to keep track of what links we are currently working with in the linksToDuplicate vector
				int currentLink = 0;
				//create genes for the new nodes
				//start at 1 since the 0 node is the input node and don't index the last node as it is the output node
				for(int node = 1; node < nodesToDuplicate.size()-1; node++)
				{
					//single duplicate node case
					if(nodesToDuplicate.size() == 3)
					{	
						//1. Create the new innovation for the new neuron
						
						//First grab the new innovation ID
						int idNeuron1 = innovationList.getNextInnovationID();
						
						//Identify the neurons this neuron is between
						int from = nodesToDuplicate.elementAt(node-1).getNeuronID();
						int to = nodesToDuplicate.elementAt(nodesToDuplicate.size()-1).getNeuronID(); 
						
						//just duplicate the splitX and splitY values, they are virtually not used
						//might consider removing these in a future revision of this system.
						double newDepth = nodesToDuplicate.elementAt(node).getSplitY();
						double newWidth = nodesToDuplicate.elementAt(node).getSplitX();
						
						//Innovation type, 0 = node, 1 = link
						//Node type, 0=none, 1=input, 2=output, 3=hidden, 4=bias
						int newNeuronID = innovationList.createNewInnovation(from, to, 0, 3, newWidth, newDepth);
						
						//2. Create a new gene
						//start by creating a variation operator for randomisation
						//mutation rate of this operator should be SDSubMutationRate and not the standard Mutaiton Rate
						VariationOperator_Element_NEAT_Gene op = new VariationOperator_Element_NEAT_Gene("initialization operator",
																										((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																										((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																										((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxWeightPerturbation(),
																										((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																										((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxActivationPerturbation());
						
						//VariationOperator_Element_NEAT_Gene op2 = new VariationOperator_Element_NEAT_Gene("initialization operator - does nothing");
						
						Element_NEAT_NGene newNode = new Element_NEAT_NGene("segmental duplication node SNC",
																		 genome, //owner is this genome
																		 op, //initialise with this operation
																		 idNeuron1, //new innovation id
																		 false, //is not recurrent ?
																		 3, //neuron type is hidden
																		 nodesToDuplicate.elementAt(node).getActivationResponse(), //activation response for new node
																		 newWidth, //splitX
																		 newDepth, //splitY
																		 newNeuronID //new unique neuronID
																		 );
						genome.addNeuron(newNode);
						
						//3. Create the first new link
						int idLink1 = innovationList.getNextInnovationID();
						//create the new innovation for the first link
						innovationList.createNewInnovation(from, newNeuronID, 1);
						//create the new link gene
						Element_NEAT_LGene nLink1 = new Element_NEAT_LGene("segmental duplication link for node SNC",
								  genome,
								  op, //this initialisation does nothing
								  idLink1, //this is getting the next neuron id it should be innovation id
								  false, //not recurrent?
								  from, //from the old neuron
								  newNeuronID, //to the new neuron
								  linksToDuplicate.elementAt(0).getLinkWeight()); //set this vector weight the original value (0 index due to special case)
						genome.addLink(nLink1);
						
						//4. Create the second new link
						int idLink2 = innovationList.getNextInnovationID();
						//create the new innovation for the first link
						innovationList.createNewInnovation(newNeuronID, to, 1);
						//create the new link gene
						Element_NEAT_LGene nLink2 = new Element_NEAT_LGene("segmental duplication link for node SNC",
								  genome,
								  op, //this initialisation does nothing
								  idLink2, //this is getting the next neuron id it should be innovation id
								  false, //not recurrent?
								  newNeuronID, //from the old neuron
								  to, //to the new neuron
								  linksToDuplicate.elementAt(1).getLinkWeight()); //set this vector weight to the original value (1 index due to special case)
						genome.addLink(nLink2);
					}
					//The N nodes case
					else
					{
						//first hidden neuron case
						if(node == 1)
						{
							//1. Create the new innovation for the new neuron
							
							//First grab the new innovation ID
							int innovationIDForNewNeuron = innovationList.getNextInnovationID();
							//Identify the neurons this neuron is between
							int from = nodesToDuplicate.elementAt(0).getNeuronID();
							int to = innovationList.getNextNeuronID()+1; //this is the new neuron ID for the next new neuron not this current new neuron.
							
							//just duplicate the splitX and splitY values, they are virtually not used
							//might consider removing these in a future revision of this system.
							double newDepth = nodesToDuplicate.elementAt(node).getSplitY();
							double newWidth = nodesToDuplicate.elementAt(node).getSplitX();
							
							//Innovation type, 0 = node, 1 = link
							//Node type, 0=none, 1=input, 2=output, 3=hidden, 4=bias
							int newNeuronID = innovationList.createNewInnovation(from, to, 0, 3, newWidth, newDepth);
							
							//2. Create a new gene
							//start by creating a variation operator for randomisation
							//mutation rate of this operator should be SDSubMutationRate and not the standard Mutaiton Rate
							VariationOperator_Element_NEAT_Gene op = new VariationOperator_Element_NEAT_Gene("initialization operator",
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxWeightPerturbation(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxActivationPerturbation());
							
							Element_NEAT_NGene newNode = new Element_NEAT_NGene("segmental duplication node MNC Head",
																			 genome, //owner is this genome
																			 op, //initialise with this operation
																			 innovationIDForNewNeuron, //new innovation id
																			 false, //is not recurrent ?
																			 3, //neuron type is hidden
																			 nodesToDuplicate.elementAt(node).getActivationResponse(), //activation response for new node
																			 newWidth, //splitX
																			 newDepth, //splitY
																			 newNeuronID //new unique neuronID
																			 );
							genome.addNeuron(newNode);
							
							//3. Create the first new link
							int idLink1 = innovationList.getNextInnovationID();
							//create the new innovation for the first link
							innovationList.createNewInnovation(from, newNeuronID, 1);
							//create the new link gene
							Element_NEAT_LGene nLink1 = new Element_NEAT_LGene("segmental duplication link for node MNC Head",
									  genome,
									  op, //this initialisation causes potential mutation to this link
									  idLink1, //this is getting the next neuron id it should be innovation id
									  false, //not recurrent?
									  from, //from the old neuron
									  newNeuronID, //to the new neuron
									  linksToDuplicate.elementAt(currentLink).getLinkWeight()); //set this vector weight the original value
							genome.addLink(nLink1);
							//done handling the links, increment the index
							currentLink++;
						}
						else if(node > 1 && nodesToDuplicate.elementAt(node+1).getNeuronType() != 2)
						{
							//1. Create the new innovation for the new neuron
							
							//First grab the new innovation ID
							int innovationIDForNewNeuron = innovationList.getNextInnovationID();
							//Identify the neurons this neuron is between
							int from = innovationList.getNextNeuronID()-1; //this is the neuron ID of the last new neuron
							int to = innovationList.getNextNeuronID()+1; //this is the new neuron ID for the next new neuron not this current new neuron.
							
							//just duplicate the splitX and splitY values, they are virtually not used
							//might consider removing these in a future revision of this system.
							double newDepth = nodesToDuplicate.elementAt(node).getSplitY();
							double newWidth = nodesToDuplicate.elementAt(node).getSplitX();
							
							//Innovation type, 0 = node, 1 = link
							//Node type, 0=none, 1=input, 2=output, 3=hidden, 4=bias
							int newNeuronID = innovationList.createNewInnovation(from, to, 0, 3, newWidth, newDepth);
							
							//2. Create a new gene
							//start by creating a variation operator for randomisation
							//mutation rate of this operator should be SDSubMutationRate and not the standard Mutaiton Rate
							VariationOperator_Element_NEAT_Gene op = new VariationOperator_Element_NEAT_Gene("initialization operator",
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxWeightPerturbation(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxActivationPerturbation());
							
							Element_NEAT_NGene newNode = new Element_NEAT_NGene("segmental duplication node MNC Mid",
																			 genome, //owner is this genome
																			 op, //initialise with this operation
																			 innovationIDForNewNeuron, //new innovation id
																			 false, //is not recurrent ?
																			 3, //neuron type is hidden
																			 nodesToDuplicate.elementAt(node).getActivationResponse(), //activation response for new node
																			 newWidth, //splitX
																			 newDepth, //splitY
																			 newNeuronID //new unique neuronID
																			 );
							genome.addNeuron(newNode);
							
							//3. Create the first new link
							int idLink1 = innovationList.getNextInnovationID();
							//create the new innovation for the first link
							innovationList.createNewInnovation(from, newNeuronID, 1);
							//create the new link gene
							Element_NEAT_LGene nLink1 = new Element_NEAT_LGene("segmental duplication link for node MNC Mid",
									  genome,
									  op, //this initialisation causes potential mutation to this link
									  idLink1, //this is getting the next neuron id it should be innovation id
									  false, //not recurrent?
									  from, //from the old neuron
									  newNeuronID, //to the new neuron
									  linksToDuplicate.elementAt(currentLink).getLinkWeight()); //set this vector weight the original value
							genome.addLink(nLink1);
							//done handling the links, increment the index
							currentLink++;
						}
						//handle the last case where we have to add the link before this node
						//and the link to the output node.
						else
						{
							//1. Create the new innovation for the new neuron
							
							//First grab the new innovation ID
							int innovationIDForNewNeuron = innovationList.getNextInnovationID();
							//Identify the neurons this neuron is between
							int from = innovationList.getNextNeuronID()-1; //this is the neuron ID of the last new neuron
							int to = nodesToDuplicate.elementAt(node+1).getNeuronID(); //this is the new neuron ID for the next new neuron not this current new neuron.
							
							//just duplicate the splitX and splitY values, they are virtually not used
							//might consider removing these in a future revision of this system.
							double newDepth = nodesToDuplicate.elementAt(node).getSplitY();
							double newWidth = nodesToDuplicate.elementAt(node).getSplitX();
							
							//Innovation type, 0 = node, 1 = link
							//Node type, 0=none, 1=input, 2=output, 3=hidden, 4=bias
							int newNeuronID = innovationList.createNewInnovation(from, to, 0, 3, newWidth, newDepth);
							
							//2. Create a new gene
							//start by creating a variation operator for randomisation
							//mutation rate of this operator should be SDSubMutationRate and not the standard Mutaiton Rate
							VariationOperator_Element_NEAT_Gene op = new VariationOperator_Element_NEAT_Gene("initialization operator",
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxWeightPerturbation(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getSubMutationRateSD(),
																											((ParameterSet_Evolution_SDNEAT)genome.getTemplate()).getMaxActivationPerturbation());
							
							Element_NEAT_NGene newNode = new Element_NEAT_NGene("segmental duplication node MNC Tail",
																			 genome, //owner is this genome
																			 op, //initialise with this operation
																			 innovationIDForNewNeuron, //new innovation id
																			 false, //is not recurrent ?
																			 3, //neuron type is hidden
																			 nodesToDuplicate.elementAt(node).getActivationResponse(), //activation response for new node
																			 newWidth, //splitX
																			 newDepth, //splitY
																			 newNeuronID //new unique neuronID
																			 );
							genome.addNeuron(newNode);
							
							//3. Create the first new link
							int idLink1 = innovationList.getNextInnovationID();
							//create the new innovation for the first link
							innovationList.createNewInnovation(from, newNeuronID, 1);
							//create the new link gene
							Element_NEAT_LGene nLink1 = new Element_NEAT_LGene("segmental duplication link for node MNC Tail",
									  genome,
									  op, //this initialisation does nothing
									  idLink1, //this is getting the next neuron id it should be innovation id
									  false, //not recurrent?
									  from, //from the old neuron
									  newNeuronID, //to the new neuron
									  linksToDuplicate.elementAt(currentLink).getLinkWeight()); //set this vector weight the original value (0 index due to special case)
							genome.addLink(nLink1);
							
							//4. Create the second new link
							int idLink2 = innovationList.getNextInnovationID();
							//create the new innovation for the first link
							innovationList.createNewInnovation(newNeuronID, to, 1);
							//create the new link gene
							Element_NEAT_LGene nLink2 = new Element_NEAT_LGene("segmental duplication link for node MNC Tail",
									  genome,
									  op, //this initialisation does nothing
									  idLink2, //this is getting the next neuron id it should be innovation id
									  false, //not recurrent?
									  newNeuronID, //from the old neuron
									  to, //to the new neuron
									  linksToDuplicate.elementAt(currentLink+1).getLinkWeight()); //set this vector weight to the original value (1 index due to special case)
							genome.addLink(nLink2);
						}//end the N-case IF statement
					}//end the IF statement separating the 1 node special case from the N node case
				}//end the FOR loop iterating through the nodes for duplication.
			}//end the IF statement that handles actual segemental duplication
		}//end the initial IF statement stopping execution if there is no hidden nodes
		else
		{
			return;
		}
	}
	
	private Element_NEAT_LGene getPathNextRandomNode(Element_NEAT_NGene fromNeuron, Individual_NEATIndividual genome)
	{
		//vector to contain all the locations in the genomes link array of links from this Neuron
		Vector<Integer> linkLocations = new Vector<Integer>();
		//find all the outward links that are not recurrent and are enabled
		for(int link = 0; link < genome.getNumLinks(); link++)
		{
			if(genome.getLinkAt(link).getFromNeuron() == fromNeuron.getNeuronID() &&
					!genome.getLinkAt(link).getRecurrent() &&
					genome.getLinkAt(link).isEnabled())
				{
					linkLocations.add(link);
				}
		}
		
		//return a random link if there is more than one, or return the sole link if there is only one
		int linkToReturn = 0;
		if(linkLocations.size() > 1)
		{
			linkToReturn = linkLocations.elementAt(randInt(0,linkLocations.size()-1));
		}
		else
		{
			linkToReturn = linkLocations.elementAt(0);
		}
		
		return genome.getLinkAt(linkToReturn);
	}
	
	//tests to see if the given neuron id exists in the given genome
	//returns true if it exists
	private boolean alreadyHaveThisNeuronID(int id, Individual_NEATIndividual genome)
	{
		for(int neuron = 0; neuron < genome.getNumNeurons(); neuron++)
		{
			if(id == genome.getNeuronAt(neuron).getNeuronID())
			{
				return true;
			}
		}
		return false;
	}
	
	//given a neuron ID this function finds its position in the vector of neurons
	private int getElementPosition(int neuronId, Individual_NEATIndividual genome)
	{
		for(int neuron = 0; neuron < genome.getNumNeurons(); neuron++)
		{
			if(genome.getNeuronAt(neuron).getNeuronID() == neuronId)
			{
				return neuron;
			}
		}
		Display.error(""+this.getClass().getName()+"::mutate(.) - Attempting to a link to a node which apparently doesn't exist in this genome");
		return -1;
	}
	
	//returns a random integer between x and y
	private int randInt(int x, int y)
	{
		return rand.nextInt(Integer.MAX_VALUE)%(y-x+1)+x;
	}

}
