/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.ext.evolution;

import java.util.*;

import simbad.picoevo.core.evolution.*;
import simbad.picoevo.core.representation.*;
import simbad.picoevo.ext.representation.*;
import simbad.picoevo.toolbox.*;

public class VariationOperator_Population_NEAT_CrossOver extends VariationOperator {

	//create a random number generator
	Random rand;
	
	public VariationOperator_Population_NEAT_CrossOver(String __name) {
		super(__name);
		rand = new Random();
	}

	public VariationOperator_Population_NEAT_CrossOver(String __name, double rate) {
		super(__name, rate);
		rand = new Random();
	}

	@Override
	public void applyOperator(EvolutionObject object) {
		applyOperator(object,null,null);
	}
	
	public Individual_NEATIndividual applyOperator(EvolutionObject object, Individual_NEATIndividual mother,
							  Individual_NEATIndividual father)
	{
		if(mother == null && father == null)
		{
			Display.error(""+this.getClass().getName()+"::applyOperator(.) - NEAT doesn't apply CrossOver across the entire population this way, this should not be called.");
			return null;
		}
		else
		{
			return performCrossOver(object, mother, father);
		}
	}
	
	public Individual_NEATIndividual performCrossOver(EvolutionObject object, Individual_NEATIndividual mother, Individual_NEATIndividual father)
	{
		//The current population
		Population_NEATPopulation population = (Population_NEATPopulation)object;
		//for readability in the code
		int parentTypeMom = 0; 
		int parentTypeDad = 1;
		
		//1. Calculate which genome to use the disjoint/excess genes from. This is the fittest genome.
		int bestParent = -1;
		
		//If they are of equal fitness use the shorter (we want to keep the networks ass small as possible)
		if(mother.getAdjustedFitness() == father.getAdjustedFitness())
		{
			//If they are equal fitness and length just choose one at random
			if(mother.getNumLinks() == father.getNumLinks())
			{
				bestParent = randInt(0,1);
			}
			else
			{
				if(mother.getNumLinks() < father.getNumLinks())
				{
					bestParent = parentTypeMom;
				}
				else
				{
					bestParent = parentTypeDad;
				}
			}
		}
		else
		{
			if(mother.getAdjustedFitness() > father.getAdjustedFitness())
			{
				bestParent = parentTypeMom;
			}
			else
			{
				bestParent = parentTypeDad;
			}
		}
		
		//These vectors will hold the offsprings nodes and genes
		Vector<Element_NEAT_NGene> offspringNeurons = new Vector<Element_NEAT_NGene>();
		Vector<Element_NEAT_LGene> offspringLinks = new Vector<Element_NEAT_LGene>();
		
		//Temporary vector to store all added node ID's
		Vector<Integer> vecNeurons = new Vector<Integer>();
		
		//Integers to keep track of each parents gene we are on
		int currentMom = 0;
		int currentDad = 0;
		
		//This will hold copies of the genes we wish to add at each step
		Element_NEAT_LGene selectedLinkGene = null;
		
		//Step through each parents genes until we reach the end of both.
		while (!((currentMom == mother.getNumLinks()) && (currentDad == father.getNumLinks())))
		{
			//the end of the mothers genes have been reached
			if ((currentMom == mother.getNumLinks()) && (currentDad != father.getNumLinks()))
			{
				//if dad is the best fit
				if (bestParent == parentTypeDad)
				{
					//add dads genes
					selectedLinkGene = father.getLinkAt(currentDad);
				}
				++currentDad;
			}
			//the end of dad's genes have been reached
			else if ((currentDad == father.getNumLinks()) && (currentMom != mother.getNumLinks()))
			{
				//if mom is the best fit
				if( bestParent == parentTypeMom)
				{
					//add moms genes
					selectedLinkGene = mother.getLinkAt(currentMom);
				}
				++currentMom;
			}
			//if moms innovation number is less than dads
			else if ((mother.getLinkAt(currentMom).getInnovationID()) < (father.getLinkAt(currentDad).getInnovationID()))
			{
				//if mom is the best fit
				if( bestParent == parentTypeMom)
				{
					selectedLinkGene = mother.getLinkAt(currentMom);
				}
				++currentMom;
			}
			//if dads innovation number is less than moms
			else if ((father.getLinkAt(currentDad).getInnovationID()) < (mother.getLinkAt(currentMom).getInnovationID()))
			{
				//if dad is the best fit
				if( bestParent == parentTypeDad)
				{
					selectedLinkGene = father.getLinkAt(currentDad);
				}
				++currentDad;
			}
			//if innovation numbers are the same
			else if ((father.getLinkAt(currentDad).getInnovationID()) == (mother.getLinkAt(currentMom).getInnovationID()))
			{
				//grab a gene from either parent
				if(Math.random() < 0.5)
				{
					selectedLinkGene = mother.getLinkAt(currentMom);
				}
				else
				{
					selectedLinkGene = father.getLinkAt(currentDad);
				}
				//move onto the next gene of each parent
				++currentDad;
				++currentMom;
			}
			else
			{
				//Display.info("error state");
				//Display.error(""+this.getClass().getName()+"::CrossOver - Something has gone horribly wrong");
			}
			
			//add the selected gene if not already added
			if (selectedLinkGene != null)
			{
				if (offspringLinks.size() == 0)
				{
					offspringLinks.add(selectedLinkGene);
				}
				else
				{
					//Might want to double check this, I think the ordering of the nodes is important here
					//and our nodes may not be in the right order
					if(offspringLinks.get(offspringLinks.size()-1).getInnovationID() !=
						selectedLinkGene.getInnovationID())
					{
						offspringLinks.add(selectedLinkGene);
					}
				}
				
				//Check if we already have the nodes referred to in selectedLinkGene?
				//If not, they may need to be added
				addNeuronID(selectedLinkGene.getFromNeuron(), vecNeurons);
				addNeuronID(selectedLinkGene.getToNeuron(), vecNeurons);
			}
		}
		
		//now create the required nodes. First sort them into order
		Collections.sort(vecNeurons);
		
		for(int i = 0; i < vecNeurons.size(); i++)
		{
			//NOTE: These genes do not have their owner set, they aren't an individual yet.
			//The ownership of these nodes will be corrected by the Individual constructor
			//A new individual MUST have its nodes constructed or be a complete clone.
			offspringNeurons.add((population.innovationList).createNeuronFromID(i));
		}
		
		//finally, create the genome
		Individual_NEATIndividual offspring = new Individual_NEATIndividual("("+mother._name+"-"+father._name+")", 
																			null, //this offsprings population 
																			population.nextGenomeID, //the new unique id for this individual
																			offspringNeurons,//new neurons
																			offspringLinks,//new links
																			0, //depth of the network??? 
																			mother.numInputs(), //the number of inputs
																			mother.numOutputs(), //the number of outputs
																			-1 //genome species, for record keeping, this is not speciated yet
																			);
		return offspring;
		
	}
	
	//Checks to see if a specific node id has been added to a vector of nodes
	//If not, then the new ID gets added.
	private void addNeuronID(int nodeID, Vector<Integer> nodeIDs)
	{
		for (int i = 0; i < nodeIDs.size(); i++)
		{
			if(nodeIDs.get(i) == nodeID)
			{
				return;
			}
		}
		nodeIDs.add(nodeID);
		return;
	}
	
	//returns a random integer between x and y
	private int randInt(int x, int y)
	{
		return rand.nextInt(Integer.MAX_VALUE)%(y-x+1)+x;
	}

}
