/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.ext.representation;

import java.util.*;

import simbad.picoevo.ext.evolution.*;
import simbad.picoevo.toolbox.*;

public class Population_NEATPopulation_InnovationList {
	
	protected Vector<Innovation> innovations;
	protected int nextNeuronID;
	protected int nextInnovationID;
	
	public Population_NEATPopulation_InnovationList()
	{
		nextNeuronID = 0;
		nextInnovationID = 0;
		innovations = new Vector<Innovation>();
	}
	
	public void displayInformation()
	{
		Display.info("###########################");
		Display.info(" ( \""+"Innovation List"+"\",{");
		Display.info("###########################");
		for(int i = 0; i < innovations.size(); i++)
		{
			Display.info_nocr("Innovation Record "+i+":");
			innovations.get(i).displayInformation();
			Display.info("end "+i);
		}
		Display.info_nocr(" nextNeuronID :"+nextNeuronID);
		Display.info_nocr(" nextInnovationID :"+nextInnovationID);
		Display.info(" }");
	}
	
	public String toString()
	{
		StringBuffer output = new StringBuffer();
		for(int i = 0; i < innovations.size(); i++)
		{
			output.append(innovations.get(i).toString());
			output.append("\r\n");
		}
		return output.toString();
	}
	
	public Population_NEATPopulation_InnovationList(Vector<Element_NEAT_LGene> start_links,
													Vector<Element_NEAT_NGene> start_neurons)
	{
		nextNeuronID = 0;
		nextInnovationID = 0;
		
		//add the neurons to the innovation list
		for(int node = 0; node < start_neurons.size(); node++)
		{
			innovations.add(new Innovation(start_neurons.get(node), nextInnovationID, nextNeuronID));
			nextNeuronID++;
			nextInnovationID++;
		}
		
		for(int link = 0; link < start_links.size(); link++)
		{
			//Innovation type, 0 = node, 1 = link;
			innovations.add(new Innovation(start_links.get(link).fromNeuron,
										   start_links.get(link).toNeuron,
										   1,
										   nextInnovationID));
			nextInnovationID++;
		}
	}
	
	/**This function will initialize an innovation list that hasn't been initialized upon creation
	 * only call this if the innovation list has been constructed without an initial genome */
	public void initializeInnovationList(Vector<Element_NEAT_LGene> given_links,
										 Vector<Element_NEAT_NGene> given_neurons)
	{
		Vector<Element_NEAT_LGene> start_links = given_links;
		Vector<Element_NEAT_NGene> start_neurons = given_neurons;
		
		if(innovations.size() == 0)
		{
			nextNeuronID = 0;
			nextInnovationID = 0;

			//add the neurons to the innovation list
			for(int node = 0; node < start_neurons.size(); node++)
			{
				innovations.add(new Innovation(start_neurons.get(node), nextInnovationID, nextNeuronID));
				nextNeuronID++;
				nextInnovationID++;
			}

			for(int link = 0; link < start_links.size(); link++)
			{
				//Innovation type, 0 = node, 1 = link;
				innovations.add(new Innovation(start_links.get(link).fromNeuron,
						start_links.get(link).toNeuron, 1, nextInnovationID));
				nextInnovationID++;
			}
		}
		else
		{
			Display.error(""+this.getClass().getName()+"::initialise(.) - Attempting to initialize an already initialized Innovation List");
		}
	}
	
	/**Checks if this innovation has already occured, 
	 * if it has it returns the innovation ID, 
	 * if not it returns a negative value.
	 * Innovation type, 0 = node, 1 = link;*/
	public int checkInnovation(int in, int out, int innoType)
	{
		//iterate through the innovations looking for a match on all the parameters
		for(int inno = 0; inno < innovations.size(); inno++)
		{
			if(innovations.get(inno).neuronIn == in &&
			   innovations.get(inno).neuronOut == out &&
			   innovations.get(inno).innovationType == innoType)
			{
				return innovations.get(inno).innovationID;
			}
		}
		//if no match, return
		return -1;
	}
	
	/**Creates a new innovation and returns its NEURON ID
	 * Innovation type, 0 = node, 1 = link;*/
	public int createNewInnovation(int in, int out, int innoType)
	{
		Innovation newInno = new Innovation(in,out,innoType,nextInnovationID);
		
		//Innovation type, 0 = node, 1 = link;
		if(innoType == 0)
		{
			newInno.neuronID = nextNeuronID;
			nextNeuronID++;
		}
		innovations.add(newInno);
		nextInnovationID++;
		//return -1 if a link
		if(innoType == 0)
			return nextNeuronID-1;
		else
			return -1;
	}
	
	/**Creates a new innovation and returns its NEURON ID
	 * Innovation type, 0 = node, 1 = link;*/
	public int createNewInnovation(int from, int to, int innoType, int neuronType, double sX, double sY)
	{
		Innovation newInno = new Innovation(innoType, nextInnovationID, from, to, nextNeuronID, neuronType, sX, sY);
		
		//Innovation type, 0 = node, 1 = link;
		if(innoType == 0)
		{
			newInno.neuronID = nextNeuronID;
			nextNeuronID++;
		}
		innovations.add(newInno);
		nextInnovationID++;
		//return -1 if a link
		if(innoType == 0)
			return nextNeuronID-1;
		else
			return -1;
	}
	
	/**Given a neuronID this function returns a clone of that neuron*/
	public Element_NEAT_NGene createNeuronFromID(int neuronID)
	{
		VariationOperator_Element_NEAT_Gene op = new VariationOperator_Element_NEAT_Gene("initialization operator - does nothing");
		Element_NEAT_NGene node;
		
		for(int i = 0; i < innovations.size(); i++)
		{
			if(innovations.get(i).neuronID == neuronID)
			{
				node = new Element_NEAT_NGene("node", null, //owner is this genome
											  op, //initialise with this operation
											  innovations.get(i).innovationID, //new innovation id
											  false, //is not recurrent ?
											  innovations.get(i).neuronType, //neuron type is hidden
											  0.0, //activation response for new node
											  innovations.get(i).splitX, //splitX
											  innovations.get(i).splitY, //splitY
											  innovations.get(i).neuronID //new unique neuronID
											  );
				return node;
			}
		}
		node = new Element_NEAT_NGene("INVALID node from crossover INVALID", null, //owner is this genome
				  					  op, //initialise with this operation
				  					  0, //new innovation id
				  					  false, //is not recurrent ?
				  					  0, //neuron type is hidden
				  					  0, //activation response for new node
				  					  0, //splitX
				  					  0, //splitY
				  					  0 //new unique neuronID
				  					  );
		return node;
	}
	
	public Population_NEATPopulation_InnovationList.Innovation get(int index)
	{
		return innovations.get(index);
	}
	
	public void add(Population_NEATPopulation_InnovationList.Innovation element)
	{
		this.innovations.add(element);
	}

	public int getNextNeuronID() {
		return nextNeuronID;
	}

	public void setNextNeuronID(int nextNeuronID) {
		this.nextNeuronID = nextNeuronID;
	}

	public int getNextInnovationID() {
		return nextInnovationID;
	}

	public void setNextInnovationID(int nextInnovationID) {
		this.nextInnovationID = nextInnovationID;
	}
	
	public int getNeuronID(int id)
	{
		for (int i = 0; i < innovations.size(); i++)
		{
			if(innovations.elementAt(i).innovationID == id)
			{
				return innovations.elementAt(i).neuronID;
			}
		}
		return -1;
	}
	
	public int size()
	{
		return innovations.size();
	}
	
	protected class Innovation 
	{
		//Innovation type, 0 = node, 1 = link;
		int innovationType= 0;
		int innovationID = 0;
		int neuronIn = 0;
		int neuronOut = 0;
		int neuronID = 0;
		//Neuron Type, 0=none, 1=input, 2=output, 3=hidden, 4=bias
		int neuronType = 0;
		//If it is a node this information is kept for display purposes (may be able to remove this)
		double splitX, splitY;
		
		public Innovation()
		{
			neuronIn = -1;
			neuronOut = -1;
			innovationType = -1;
			innovationID = -1;
			neuronID = -1;
			splitX = 0;
			splitY = 0;
			neuronType = -1; //none aka not a neuron
		}
		
		public void displayInformation()
		{
			Display.info(" ( \""+"innovation record"+"\",{");
			Display.info_nocr(" neuronIn :"+neuronIn);
			Display.info_nocr(" neuronOut :"+neuronOut);
			Display.info_nocr(" innovationType :"+innovationType);
			Display.info_nocr(" innovationID :"+innovationID);
			Display.info_nocr(" neuronID :"+neuronID);
			Display.info_nocr(" splitX :"+splitX);
			Display.info_nocr(" splitY :"+splitY);
			Display.info_nocr(" neuronType :"+neuronType);
			Display.info(" }");
		}
		
		public String toString()
		{
			//Innovation type, 0 = node, 1 = link;
			String iType = new String();
			
			if(innovationType == 0)
				iType = "node";
			else if(innovationType == 1)
				iType = "link";
			
			//0=none, 1=input, 2=output, 3=hidden, 4=bias
			String nType = new String();
			
			if(neuronType == 0)
				nType = "none";
			else if(neuronType == -1)
				nType = "not a neuron";
			else if(neuronType == 1)
				nType = "input";
			else if(neuronType == 2)
				nType = "output";
			else if(neuronType == 3)
				nType = "hidden";
			else if(neuronType == 4)
				nType = "bias";

			StringBuffer output = new StringBuffer();
			output.append("iID: ");
			output.append(innovationID);
			output.append(" ");
			output.append(iType);
			output.append(" ->N: ");
			output.append(neuronIn);
			output.append(" N->: ");
			output.append(neuronOut);
			output.append(" ");
			output.append(nType);
			return output.toString();
		}
		
		public Innovation(int in, int out, int innoType, int innoID)
		{
			neuronIn = in;
			neuronOut = out;
			innovationType = innoType;
			innovationID = innoID;
			neuronID = -1;
			splitX = 0;
			splitY = 0;
			neuronType = 0; //none aka not a neuron
		}
		
		public Innovation(Element_NEAT_NGene neuron, int innoID, int nID)
		{
			innovationID = innoID;
			neuronID = nID;
			innovationType = 0; //obviously a new neuron
			splitX = neuron.splitX;
			splitY = neuron.splitY;
			neuronType = neuron.neuronType;
			neuronIn = -1;
			neuronOut = -1;
		}
		
		public Innovation(int innoType, int innoID, int in, int out, int nID, int nT, double sX, double sY)
		{
			innovationType= innoType;
			innovationID = innoID;
			neuronIn = in;
			neuronOut = out;
			neuronID = nID;
			neuronType = nT;
			splitX = sX;
			splitY = sY;
		}
	}
}
