/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.picoevo.ext.representation;

import java.util.*;

import simbad.picoevo.core.representation.*;
import simbad.picoevo.ext.evolution.*;
import simbad.picoevo.toolbox.*;

public class Population_NEATPopulation_SpeciesList{

	public Population populationOwner;
	public Vector<Species> speciesList;
	private int numOfSpecies;
	public double compatibilityThreshold;
	
	public Population_NEATPopulation_SpeciesList(Population populationOwner)
	{
		speciesList = new Vector<Species>();
		numOfSpecies = 0;
		compatibilityThreshold = ((ParameterSet_Evolution_NEAT)populationOwner.getOwner().getTemplate()).getInitialCompatibilityThreshold();
		setPopulationSpace(populationOwner);
	}
	
	public void displayInformation()
	{
		Display.info("###########################");
		Display.info(" ( \""+"Species List"+"\",{");
		Display.info("###########################");
		Display.info(" Total Number of Species :" + numOfSpecies);
		Display.info(" ## Displaying Species ## :");
		for(int i = 0; i < speciesList.size(); i++)
		{
			Display.info("Species Index: " + i);
			speciesList.get(i).displayInformation();
			Display.info("end "+i);
		};
		Display.info(" }");
	}
	
	public String toString()
	{
		StringBuffer output = new StringBuffer();
		for(int i = 0; i < speciesList.size(); i++)
		{
			output.append("Species Index: ");
			output.append(i);
			output.append(" \r\n { \r\n ");
			output.append(speciesList.get(i).toString());
			output.append("}\r\n ");
		}
		return output.toString();
	}
	
	public void setPopulationSpace(Population populationSpace)
	{
		populationOwner = populationSpace;
	}
	
	public void speciate(ArrayList<Individual_NEATIndividual> population)
	{
		boolean knownSpecies = false;
		boolean isLeader = false;
		double compatibility = 0.0;
		double totalAdjustedFitness = 0.0;
		double averageAdjustedFitness = 0.0;
		
		//Adjust compatibility threshold here from the parameters class
		adjustCompatibilityThreshold();
		
		//iterate through each genome and speciate
		for(int gen = 0; gen < population.size(); gen++)
		{
			//calculate compatibility score with each species leader, if compatible add to species otherwise create a new species
			for(int spec = 0; ((spec < speciesList.size())); spec++)
			{
				compatibility = population.get(gen).getCompatibilityScore(speciesList.elementAt(spec).alphaMember);
				
				if(compatibility <= compatibilityThreshold)
				{
					//this individual is similar and will be added to this species
					speciesList.elementAt(spec).add(population.get(gen));
					//tell the genome that it is now a member of this species
					population.get(gen).genomeSpecies = speciesList.elementAt(spec).alphaMember.genomeSpecies;
					//identify that this is a known species
					knownSpecies = true;
					//leave the inner loop
					spec = speciesList.size();
				}
			}
			if(!knownSpecies)
			{
				//let the new leader know his species number
				population.get(gen).genomeSpecies = numOfSpecies;
				//create and add the species
				speciesList.add(new Species(this, population.get(gen),numOfSpecies++));
			}
			knownSpecies = false;
		}
		
		//now all the genomes have been assigned a species, the fitness scores need to be adjusted for
		//fitness sharing and species age.
		adjustFitnesses(population);
		
		//calculate new adjusted total and average fitness for the population
		for(int gen = 0; gen < population.size(); gen++)
		{
			totalAdjustedFitness += population.get(gen).getAdjustedFitness();
		}
		
		averageAdjustedFitness = totalAdjustedFitness/population.size();
		
		//calculate how many offspring each member of the population should spawn
		double toSpawn = 0.0;
		for(int gen = 0; gen < population.size(); gen++)
		{
			toSpawn = population.get(gen).getAdjustedFitness() / averageAdjustedFitness;
			population.get(gen).setAmountToSpawn(toSpawn);
		}
		
		//Allow fast lookup of the appropriate Genome's index by GID
		HashMap<Integer, Integer> gIDToPopulationIndex = new HashMap<Integer,Integer>(); 
		for(int i = 0; i < population.size(); i++)
			gIDToPopulationIndex.put(population.get(i).genomeID, i);
		
		//Add up the expected spawn amount for each individual in this species
		//to calculate the amount of offspring required for this species
		double speciesChildrenToSpawn = 0.0;
		for(int spec = 0; spec < speciesList.size(); spec++)
		{
			for(int gen = 0; gen < speciesList.get(spec).members.size(); gen++)
			{
				speciesChildrenToSpawn += population.get(gIDToPopulationIndex.get(speciesList.get(spec).members.get(gen).getGenomeID())).getAmountToSpawn();
			}
			speciesList.get(spec).setSpawnsRequired(speciesChildrenToSpawn);
			speciesChildrenToSpawn = 0.0;
		}
	}
	
	public Species getSpeciesAt(int index)
	{
		return speciesList.get(index);
	}
	
	public void eraseSpeciesAt(int index)
	{
		speciesList.removeElementAt(index);
	}
	
	public int numOfSpecies()
	{
		return speciesList.size();
	}
	
	public int maxSpeciesNumber()
	{
		return numOfSpecies;
	}
	
	protected void adjustCompatibilityThreshold()
	{
        //If this is the first generation don't adjust the threshold since the species is identical it doesn't work in this case
        if( this.getOwner().getGenerationNumber() == 0)
        {
            return;
        }
        //If the maxNumberOfSpecies < 1 then the user has turned off speciation
        if(((ParameterSet_Evolution_NEAT)( this.getOwner().getTemplate())).getMaximumNumberOfSpecies() < 1)
        {
            return;
        }
        if(speciesList.size() > ((ParameterSet_Evolution_NEAT)(this.getOwner().getTemplate())).getMaximumNumberOfSpecies())
        {
            compatibilityThreshold += ((ParameterSet_Evolution_NEAT)(this.getOwner().getTemplate())).getThresholdIncrement();
        }
        else if(speciesList.size() < 2)
        {
            compatibilityThreshold -= ((ParameterSet_Evolution_NEAT)( this.getOwner().getTemplate())).getThresholdIncrement();
        }
        //If by some fluke no speciation has occured and this function has adjusted the CT down to 0. This will always adjust it up to a suitably small number.
        if(compatibilityThreshold <= 0.0)
        {
        	compatibilityThreshold = 0.01;
        }
        return;
	}
	
	public void adjustFitnesses(ArrayList<Individual_NEATIndividual> population)
	{
		double thisFitness = 0.0;
		
		//Allow fast lookup of the appropriate Genome's index by GID
		HashMap<Integer, Integer> gIDToPopulationIndex = new HashMap<Integer,Integer>(); 
		for(int i = 0; i < population.size(); i++)
			gIDToPopulationIndex.put(population.get(i).genomeID, i);

		//Iterate through all the species
		for(int spec = 0; spec < speciesList.size(); spec++)
		{//Iterate through this species
			for(int gen = 0; gen < speciesList.get(spec).members.size(); gen++)
			{
				//Get the raw fitness of the individual in the given population by looking up the gID of the individual
				//from the members list of this species.
				thisFitness = (population.get(gIDToPopulationIndex.get(speciesList.get(spec).members.get(gen).getGenomeID()))).getRawFitness();
				//boost fitness scores here if species is young
				if(speciesList.get(spec).age < ((ParameterSet_Evolution_NEAT)this.getOwner().getOwner().getTemplate()).getYoungBonusAgeThreshold())
				{
					thisFitness *= ((ParameterSet_Evolution_NEAT)this.getOwner().getOwner().getTemplate()).getYoungFitnessBonus();
				}
				//punish older species
				if(speciesList.get(spec).age > ((ParameterSet_Evolution_NEAT)this.getOwner().getOwner().getTemplate()).getOldAgeThreshold())
				{
					thisFitness *= ((ParameterSet_Evolution_NEAT)this.getOwner().getOwner().getTemplate()).getOldAgePenalty();
				}
			
				//apply fitness sharing to adjusted fitnesses
				population.get(gIDToPopulationIndex.get(speciesList.get(spec).members.get(gen).getGenomeID())).setAdjustedFitness((thisFitness/speciesList.get(spec).members.size()));
			}
		}
	}
	
	final public Population getOwner ()
	{
		return this.populationOwner;
	}
	
	public class Species implements Comparable<Population_NEATPopulation_SpeciesList.Species>{

		public Individual_NEATIndividual alphaMember;
		public Vector<Individual_NEATIndividual> members;
		public int speciesID;
		public double bestFitness;
		public double averageFitness;
		public int gensNoImprovement;
		public int age;
		public double spawnsRequired;
		public Population_NEATPopulation_SpeciesList owner;
		private Random rand;
		
		public Species(Population_NEATPopulation_SpeciesList owner, Individual_NEATIndividual first, int firstID)
		{
			alphaMember = first;
			speciesID = firstID;
			bestFitness = alphaMember.getRawFitness();
			averageFitness = bestFitness;
			gensNoImprovement = 0;
			age = 0;
			spawnsRequired = 0;
			members = new Vector<Individual_NEATIndividual>();
			members.add(first);
			setListSpace(owner);
			rand = new Random();
		}
		
		public void displayInformation()
		{
			Display.info(" ( \""+"Species"+"\",{");
			Display.info("Species ID #: " + speciesID);
			Display.info(" Alpha Member :");
			alphaMember.displayInformation();
			Display.info_nocr(" Other Members :");
			for(int i = 0; i < members.size(); i++)
			{
				Display.info_nocr(""+i+":");
				Display.info_nocr("Genome ID: " + members.get(i).genomeID);
				Display.info("end "+i);
			}
			Display.info(" Best Fitness :"+ bestFitness);
			Display.info(" Average Fitness :"+ averageFitness);
			Display.info(" Generations without improvement :"+ gensNoImprovement);
			Display.info(" Age in Generations :"+ age);
			Display.info(" Spawns required this generation :"+ spawnsRequired);
			Display.info(" }");
		}
		
		public String toString()
		{
			StringBuffer output = new StringBuffer();
			output.append("Species: ");
			output.append(speciesID);
			output.append("\r\n    bestFit: ");
			output.append(bestFitness);
			output.append(" averageFit: ");
			output.append(averageFitness);
			output.append(" age: ");
			output.append(age);
			output.append(" genNoImp: ");
			output.append(gensNoImprovement);
			output.append(" spwnsReq: ");
			output.append(spawnsRequired);
			output.append("\r\nAlpha: \r\n");
			output.append(alphaMember.toString());
			output.append("\r\nOthers: \r\n");
			for(int i = 0; i < members.size(); i++)
			{
				output.append("Genome ID: ");
				output.append(members.get(i).genomeID);
				output.append(" -  rFit: ");
				output.append(members.get(i).getRawFitness());
				output.append(" aFit: ");
				output.append(members.get(i).getAdjustedFitness());
				output.append("\r\n");
			}
			return output.toString();
		}
		
		public void setListSpace(Population_NEATPopulation_SpeciesList space)
		{
			owner = space;
		}
		
		public void add(Individual_NEATIndividual newMember)
		{
			if(newMember.getRawFitness() > bestFitness)
			{
				bestFitness = newMember.getRawFitness();
				gensNoImprovement = 0;
				alphaMember = newMember;
			}
			members.add(newMember);
		}
		
		/** are we killing this species or just resetting it? */
		public void purge()
		{
			members.clear();
			age++;
			gensNoImprovement++;
			spawnsRequired = 0;
		}
		
		public Individual_NEATIndividual leader()
		{
			return alphaMember;
		}
		
		public Individual_NEATIndividual spawn()
		{
			if (members.size() == 1)
			{
				return members.get(0);
			}
			else
			{
				long maxIndexSize = Math.round(((ParameterSet_Evolution_NEAT)owner.getOwner().getTemplate()).getSurvivalRate() * members.size());
				int memberInThisSpecies = randInt(0,((int)maxIndexSize));	
				return members.get(memberInThisSpecies);
			}
		}
		
		public int size()
		{
			return members.size();
		}
		
		public void setSpawnsRequired(double s)
		{
			spawnsRequired = s;
		}
		
		public int compareTo(Population_NEATPopulation_SpeciesList.Species o) 
		{
			if(this.bestFitness - o.bestFitness < 0)
				return -1;
			else if(this.bestFitness - o.bestFitness > 0)
				return 1;
			else
				return 0;
		}
		
		final public Population_NEATPopulation_SpeciesList getOwner()
		{
			return this.owner;
		}
		
		//returns a random integer between x and y
		private int randInt(int x, int y)
		{
			return rand.nextInt(Integer.MAX_VALUE)%(y-x+1)+x;
		}
	}
}
