/*
 *  NS - Neuroevolutionary Solver
 *  Copyright (C) 2008 Robert Lucas
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 -----------------------------------------------------------------------------
 * $Author: Robert Lucas $ 
 * $Date: 2008/10/11 17:20:00 $
 * $Revision: 1.00 $
 */
package simbad.sim;

import javax.vecmath.Vector3d;

public class StraightGoalLoopAgent extends Agent {

    RangeSensorBelt sonars, bumpers;
	Vector3d goal;
	Vector3d startPosition;
	Vector3d currentPosition;
	Vector3d lastPosition;
	double d; //distance to goal calculated by pythagoras. 
	double clockwise; //where to turn to close on goal.
	double angle,dot,angleCurrent; //angles and guidance for turning
	double speed;
	boolean stop; //are we home?
	
    public StraightGoalLoopAgent(Vector3d position, Vector3d givenGoal, String name) {
        super(position, name);
        // Add sensors
        bumpers = RobotFactory.addBumperBeltSensor(this);
        sonars = RobotFactory.addSonarBeltSensor(this,24);
		goal = new Vector3d(givenGoal);
		startPosition = new Vector3d(position);
		currentPosition = new Vector3d();
		lastPosition = new Vector3d();
		d = 0.0;
		clockwise = 0.0;
		angle = 0.0;
		angleCurrent = 0.0;
		dot = 0.0;
		stop = false;
		speed = 0.1;
    }

    public StraightGoalLoopAgent(Vector3d position, Vector3d givenGoal, String name, double givenSpeed) {
        super(position, name);
        // Add sensors
        bumpers = RobotFactory.addBumperBeltSensor(this);
        sonars = RobotFactory.addSonarBeltSensor(this,24);
		goal = new Vector3d(givenGoal); //too precise, it is never going to randomly hit this goal.
		startPosition = new Vector3d(position);
		currentPosition = new Vector3d();
		lastPosition = new Vector3d();
		d = 0.0;
		clockwise = 0.0;
		angle = 0.0;
		angleCurrent = 0.0;
		dot = 0.0;
		stop = false;
		speed = givenSpeed;
    }
    
    /** Initialize Agent's Behavior */
    public void initBehavior() {
    // nothing particular in this case
    	//getCoords(lastPosition);
    }

    /** Perform one step of Agent's Behavior */
    public void performBehavior() {
    	// get the current position for processing on this behaviour.
    	lastPosition.set(currentPosition);
		getCoords(currentPosition); 
		
		//Calculate distance to goal using Pythagoras. 
		d = Math.sqrt(Math.pow((currentPosition.x - goal.x), 2.0)+Math.pow((currentPosition.z - goal.z), 2.0));
		
		//Calculate if the graph that is formed between the last position, current position, and goal
		//are formed in a clockwise or counter-clockwise order. 
		//If this calculation is < 0 we want to turn counter-clockwise since the angle is clockwise
		//If this calculation is > 0 we want to turn clockwise since the angle is counter-clockwise
		clockwise = ((currentPosition.z-lastPosition.z)*(goal.x-lastPosition.x))-((currentPosition.x-lastPosition.x)*(goal.z-lastPosition.z));
		
		//Calculate Angle between our current direction and the goal and between our location and the goal.
		angle = this.instantTranslation.angle(goal);
		angleCurrent = currentPosition.angle(goal);
		
		//Calculate dot product between the goal and the current direction.
		dot = this.instantTranslation.dot(goal);

		// Collision detection
		if (bumpers.oneHasHit() && !stop) 
		{
            setTranslationalVelocity(-0.1);
            setRotationalVelocity(0.1 * Math.random());
		}
        else 
        {
           // Front left obstacle ?
            if ((sonars.hasHit(0) && (sonars.hasHit(1)) && !(d < 2.0)) && !stop) 
            {
                setRotationalVelocity(-Math.PI / 4);
                setTranslationalVelocity(speed);
            }
            // Front right obstacle ?
            else if ((sonars.hasHit(0) && (sonars.hasHit(23)) && !(d < 2.0)) && !stop) 
            {
            	setRotationalVelocity(Math.PI / 4);
            	setTranslationalVelocity(speed);
            }
            // left obstacle ?
            else if ((sonars.hasHit(3) || sonars.hasHit(4) && !(d < 2.0)) && !stop) 
            {
            	setRotationalVelocity(-Math.PI / 8);
            	setTranslationalVelocity(speed);
            }
            // right obstacle ?
            else if ((sonars.hasHit(21) || sonars.hasHit(20) && !(d < 2.0)) && !stop) 
            {
            	setRotationalVelocity(Math.PI / 8);
            	setTranslationalVelocity(speed);
            }
            //No Collision Check for Goal
            else if(d < 0.5) //Pythagorean distance to goal is less than 0.5 meters, agent must be on top of it.
            {
            	Vector3d temp = new Vector3d(goal);
            	goal = startPosition;
            	startPosition = temp;
            }
            else //If not at goal, calculate distance to goal make course correction and head to goal.
            {
            	if(angleCurrent < 0.1 && angle < 0.1)
            	{
            		setTranslationalVelocity(speed);
        			setRotationalVelocity(0.0);
            	}
            	else if(dot <= 0.0) //Angle is obtuse, turn towards goal and through 90 degree mark
            	{
            		if(clockwise < 0.0)
            			setRotationalVelocity(-Math.PI / 8); //Turn Clockwise toward goal
            		else
            			setRotationalVelocity(Math.PI / 8); //Turn Counter-Clockwise toward goal
	        		setTranslationalVelocity(speed/2);
            	}
            	else if(dot > 0.0) //Angle is accute, continue turning towards goal.
            	{
            		if(clockwise < 0.0) 
            			setRotationalVelocity(-Math.PI / 8); //Turn Clockwise toward goal
            		else
            			setRotationalVelocity(Math.PI / 8); //Turn Counter-Clockwise toward goal
	        		setTranslationalVelocity(speed/2);
            	}
            	else
            	{
            		setTranslationalVelocity(speed); //get the agent moving for some readings
            	}
            }
        }//end Behaviour If	
    }//end Function performBehavior
}
